@@ -1,6 +1,6 @@
 //
 // AutoProperties.cc for pekwm
-// Copyright © 2003-2008 Claes Nästén <me@pekdon.net>
+// Copyright © 2003-2009 Claes Nästén <me@pekdon.net>
 //
 // This program is licensed under the GNU GPL.
 // See the LICENSE file for more information.
@@ -25,23 +25,20 @@ using std::list;
 using std::map;
 using std::string;
 using std::vector;
+using std::strtol;
 
 AutoProperties *AutoProperties::_instance = 0;
 
 //! @brief Constructor for AutoProperties class
 AutoProperties::AutoProperties(void)
-    : _autoproperties_mtime(0),
-#ifdef HARBOUR
-      _harbour_sort(false),
-#endif // HARBOUR
+    : _extended(false), _harbour_sort(false),
       _apply_on_start(true)
 {
 #ifdef DEBUG
-    if (_instance)
-    {
+    if (_instance) {
         cerr << __FILE__ << "@" << __LINE__ << ": "
              << "AutoProperties(" << this << ")::AutoProperties()" << endl
-             << " *** _instance allready set: " << _instance << endl;
+             << " *** _instance already set: " << _instance << endl;
     }
 #endif // DEBUG
     _instance = this;
@@ -75,6 +72,11 @@ AutoProperties::AutoProperties(void)
     _property_map["FOCUSNEW"] = AP_FOCUS_NEW;
     _property_map["FOCUSABLE"] = AP_FOCUSABLE;
     _property_map["CFGDENY"] = AP_CFG_DENY;
+    _property_map["ALLOWEDACTIONS"] = AP_ALLOWED_ACTIONS;
+    _property_map["DISALLOWEDACTIONS"] = AP_DISALLOWED_ACTIONS;
+#ifdef OPACITY
+    _property_map["OPACITY"] = AP_OPACITY;
+#endif // OPACITY
 
     // group properties
     _group_property_map[""] = AP_NO_PROPERTY;
@@ -108,7 +110,7 @@ bool
 AutoProperties::load(void)
 {
     string cfg_file(Config::instance()->getAutoPropsFile());
-    if (! Util::requireReload(_autoproperties_path, cfg_file, _autoproperties_mtime)) {
+    if (! Util::requireReload(_cfg_state, cfg_file)) {
         return false;
     }
 
@@ -116,14 +118,23 @@ AutoProperties::load(void)
     unload();
 
     CfgParser a_cfg;
-    if (! a_cfg.parse(cfg_file)) {
+    if (! a_cfg.parse(cfg_file, CfgParserSource::SOURCE_FILE, false)) {
         cfg_file = SYSCONFDIR "/autoproperties";
-        if (! a_cfg.parse (cfg_file)) {
+        if (! a_cfg.parse (cfg_file, CfgParserSource::SOURCE_FILE, false)) {
           setDefaultTypeProperties();
           return false;
         }
     }
 
+    // Setup template parsing if requested
+    loadRequire(a_cfg, cfg_file);
+
+    if (a_cfg.is_dynamic_content()) {
+        _cfg_state.clear();
+    } else {
+        _cfg_state = a_cfg.get_file_list();
+    }
+
     // reset values
     _apply_on_start = true;
 
@@ -141,10 +152,8 @@ AutoProperties::load(void)
             parseDecorProperty(*it);
         } else if (*(*it) == "TYPERULES") {
             parseTypeProperty(*it);
-#ifdef HARBOUR
         } else if (*(*it) == "HARBOUR") {
             parseDockAppProperty(*it);
-#endif // HARBOUR
         } else if (*(*it) == "WORKSPACE") { // Workspace section
             CfgParser::Entry *workspace = (*it)->get_section();
             CfgParser::Entry *value = workspace->find_entry("WORKSPACE");
@@ -173,6 +182,33 @@ AutoProperties::load(void)
     return true;
 }
 
+/**
+ * Load autoproperties quirks.
+ */
+void
+AutoProperties::loadRequire(CfgParser &a_cfg, std::string &file)
+{
+    CfgParser::Entry *section;
+
+    // Look for requires section, 
+    section = a_cfg.get_entry_root()->find_section("REQUIRE");
+    if (section) {
+        list<CfgParserKey*> key_list;
+
+        key_list.push_back(new CfgParserKeyBool("TEMPLATES", _extended, false));
+        section->parse_key_values(key_list.begin(), key_list.end());
+        for_each(key_list.begin(), key_list.end(), Util::Free<CfgParserKey*>());
+
+        // Re-load configuration with templates enabled.
+        if (_extended) {
+            a_cfg.clear(true);
+            a_cfg.parse(file, CfgParserSource::SOURCE_FILE, true);
+        }
+    } else {
+        _extended = false;
+    }
+}
+
 //! @brief Frees allocated memory
 void
 AutoProperties::unload(void)
@@ -198,15 +234,13 @@ AutoProperties::unload(void)
     _decor_prop_list.clear();
 
     // remove dock app properties
-#ifdef HARBOUR
     for (it = _dock_app_prop_list.begin(); it != _dock_app_prop_list.end(); ++it) {
         delete *it;
     }
     _dock_app_prop_list.clear();
-#endif // HARBOUR
 
     // remove type properties
-    map<EwmhAtomName, AutoProperty*>::iterator m_it(_window_type_prop_map.begin());
+    map<AtomName, AutoProperty*>::iterator m_it(_window_type_prop_map.begin());
     for (; m_it != _window_type_prop_map.end(); ++m_it) {
         delete m_it->second;
     }
@@ -249,48 +283,52 @@ AutoProperties::findProperty(const ClassHint* class_hint,
     return 0;
 }
 
-//! @brief Parses a property match rule for window types.
-//! @param str String to parse.
-//! @param prop Property to place result in.
-//! @return true on success, else false.
-EwmhAtomName
-AutoProperties::parsePropertyMatchWindowType(const std::string &str,
-                                             Property *prop)
+/**
+ * Parse regex_str and set on regex, outputting warning with name if
+ * it fails.
+ */
+bool
+AutoProperties::parseRegexpOrWarning(RegexString &regex, const std::string regex_str, const std::string &name)
 {
-    EwmhAtomName atom = ParseUtil::getValue<EwmhAtomName>(str, _window_type_map);
-    if (atom == WINDOW_TYPE) {
-        cerr << " *** WARNING: unknown type " << str << " for autoproperty" << endl;
+    if (! regex_str.size() || regex.parse_match(Util::to_wide_str(regex_str))) {
+        return true;
+    } else {
+        cerr << " *** WARNING: invalid regexp " << regex_str << " for autoproperty " << name << endl;
+        return false;
     }
-
-    return atom;
 }
 
 //! @brief Parses a property match rule
 //! @param str String to parse.
 //! @param prop Property to place result in.
+//! @param extended Extended syntax including role and title in the name, defaults to true.
 //! @return true on success, else false.
 bool
-AutoProperties::parsePropertyMatch(const std::string &str, Property *prop)
+AutoProperties::parsePropertyMatch(const std::string &str, Property *prop, bool extended)
 {
     bool status = false;
 
     // Format of property matches are regexp,regexp . Split up in class
     // and role regexps.
     vector<string> tokens;
-    if ((Util::splitString (str, tokens, ",", 2)) == 2) {
-        // Make sure one of the two regexps compiles
-        if (prop->getHintName().parse_match (Util::to_wide_str(tokens[0]))) {
-            status = true;
-        } else {
-            cerr << " *** WARNING: failed to compile name regexp for autoproperty, "
-                 << tokens[0] << endl;
-        }
+    Util::splitString(str, tokens, ",", extended ? 5 : 2, true);
 
-        if (prop->getHintClass().parse_match (Util::to_wide_str(tokens[1]))) {
-            status = true;
-        } else {
-            cerr << " *** WARNING: failed to compile class regexp for autoproperty, "
-                 << tokens[1] << endl;
+    if (tokens.size() >= 2) {
+        // Make sure one of the two regexps compiles
+        status = parseRegexpOrWarning(prop->getHintName(), tokens[0], "name");
+        status = status && parseRegexpOrWarning(prop->getHintClass(), tokens[1], "class");
+    }
+
+    // Parse extended part of regexp, role, title and apply on
+    if (status && extended) {
+        if (status && tokens.size() > 2) {
+            status = parseRegexpOrWarning(prop->getRole(), tokens[2], "role");
+        }
+        if (status && tokens.size() > 3) {
+            status = parseRegexpOrWarning(prop->getTitle(), tokens[3], "title");
+        }
+        if (status && tokens.size() > 4) {
+            parsePropertyApplyOn(tokens[4], prop);
         }
     }
 
@@ -301,33 +339,42 @@ AutoProperties::parsePropertyMatch(const std::string &str, Property *prop)
 bool
 AutoProperties::parseProperty(CfgParser::Entry *section, Property *prop)
 {
-    vector<string> tokens;
     CfgParser::Entry *value;
 
     // Get extra matching info.
     value = section->find_entry ("TITLE");
     if (value) {
-        prop->getTitle().parse_match (Util::to_wide_str(value->get_value ()));
+        parseRegexpOrWarning(prop->getTitle(), value->get_value(), "title");
     }
     value = section->find_entry ("ROLE");
     if (value) {
-        prop->getRole().parse_match (Util::to_wide_str(value->get_value ()));
+        parseRegexpOrWarning(prop->getRole(), value->get_value(), "role");
     }
 
     // Parse apply on mask.
     value = section->find_entry ("APPLYON");
     if (value) {
-        tokens.clear();
-        if ((Util::splitString(value->get_value (), tokens, " \t", 5))) {
-            vector<string>::iterator it;
-            for (it = tokens.begin(); it != tokens.end(); ++it)
-                prop->applyAdd((uint) ParseUtil::getValue<ApplyOn>(*it, _apply_on_map));
-        }
+        parsePropertyApplyOn(value->get_value(), prop);
     }
 
     return true;
 }
 
+/**
+ * Parse property apply on.
+ */
+void
+AutoProperties::parsePropertyApplyOn(const std::string &apply_on, Property *prop)
+{
+    vector<string> tokens;
+    if ((Util::splitString(apply_on, tokens, " \t", 5))) {
+        vector<string>::iterator it(tokens.begin());
+        for (; it != tokens.end(); ++it) {
+            prop->applyAdd(static_cast<unsigned int>(ParseUtil::getValue<ApplyOn>(*it, _apply_on_map)));
+        }
+    }
+}
+
 //! @brief Parses AutopProperty
 void
 AutoProperties::parseAutoProperty(CfgParser::Entry *section, std::list<uint>* ws)
@@ -339,7 +386,7 @@ AutoProperties::parseAutoProperty(CfgParser::Entry *section, std::list<uint>* ws
     }
 
     AutoProperty* property = new AutoProperty();
-    parsePropertyMatch(section->get_value (), property);
+    parsePropertyMatch(section->get_value(), property, _extended);
 
     if (parseProperty(section, property)) {
         parseAutoPropertyValue(section, property, ws);
@@ -410,7 +457,7 @@ AutoProperties::parseTitleProperty(CfgParser::Entry *section)
         }
 
         title_property = new TitleProperty();
-        parsePropertyMatch(title_section->get_value(), title_property);
+        parsePropertyMatch(title_section->get_value(), title_property, _extended);
         if (parseProperty(title_section, title_property)) {
             CfgParser::Entry *value = title_section->find_entry("RULE");
             if (value && title_property->getTitleRule().parse_ed_s(Util::to_wide_str(value->get_value()))) {
@@ -444,7 +491,7 @@ AutoProperties::parseDecorProperty(CfgParser::Entry *section)
         }
 
         decor_property = new DecorProperty();
-        parsePropertyMatch(decor_section->get_value (), decor_property);
+        parsePropertyMatch(decor_section->get_value (), decor_property, _extended);
         if (parseProperty(decor_section, decor_property)) {
             CfgParser::Entry *value = decor_section->find_entry("DECOR");
             if (value) {
@@ -461,7 +508,6 @@ AutoProperties::parseDecorProperty(CfgParser::Entry *section)
     }
 }
 
-#ifdef HARBOUR
 /**
  * Parse dock app properties.
  */
@@ -484,7 +530,7 @@ AutoProperties::parseDockAppProperty(CfgParser::Entry *section)
         }
 
         dock_property = new DockAppProperty();
-        parsePropertyMatch(dock_section->get_value(), dock_property);
+        parsePropertyMatch(dock_section->get_value(), dock_property, _extended);
         if (parseProperty(dock_section, dock_property)) {
             CfgParser::Entry *value = dock_section->find_entry("POSITION");
             if (value) {
@@ -502,8 +548,6 @@ AutoProperties::parseDockAppProperty(CfgParser::Entry *section)
         }
     }
 }
-#endif // HARBOUR
-
 
 //! @brief Parse type auto properties.
 //! @param section Section containing properties.
@@ -513,10 +557,10 @@ AutoProperties::parseTypeProperty(CfgParser::Entry *section)
     // Get sub section
     section = section->get_section();
 
-    EwmhAtomName atom;
+    AtomName atom;
     AutoProperty *type_property;
     CfgParser::Entry *type_section;
-    map<EwmhAtomName, AutoProperty*>::iterator atom_it;
+    map<AtomName, AutoProperty*>::iterator atom_it;
 
     // Look for all type properties
     CfgParser::iterator it(section->begin());
@@ -528,7 +572,11 @@ AutoProperties::parseTypeProperty(CfgParser::Entry *section)
 
         // Create new property and try to parse
         type_property = new AutoProperty();
-        atom = parsePropertyMatchWindowType(type_section->get_value(), type_property);
+        atom = ParseUtil::getValue<AtomName>(type_section->get_value(), _window_type_map);
+        if (atom == WINDOW_TYPE) {
+            cerr << " *** WARNING: unknown type " << type_section->get_value() << " for autoproperty." << endl;
+        }
+
         if (atom != WINDOW_TYPE && parseProperty(type_section, type_property)) {
             // Parse of match ok, parse values
             parseAutoPropertyValue(type_section, type_property, 0);
@@ -562,30 +610,38 @@ AutoProperties::setDefaultTypeProperties(void)
         prop->maskAdd(AP_STICKY);
         prop->sticky = true;
         prop->maskAdd(AP_TITLEBAR);
-        prop->border = false;
+        prop->titlebar = false;
         prop->maskAdd(AP_BORDER);
         prop->border = false;
         prop->maskAdd(AP_SKIP);
         prop->skip = SKIP_MENUS|SKIP_FOCUS_TOGGLE|SKIP_SNAP|SKIP_PAGER|SKIP_TASKBAR;
         prop->maskAdd(AP_LAYER);
         prop->layer = LAYER_DESKTOP;
+        prop->maskAdd(AP_FOCUSABLE);
+        prop->focusable = false;
+        prop->maskAdd(AP_DISALLOWED_ACTIONS);
+        prop->disallowed_actions = ACTION_ACCESS_MOVE|ACTION_ACCESS_RESIZE;
 
         _window_type_prop_map[WINDOW_TYPE_DESKTOP] = prop;
     }
 
-    // DOCK  
+    // DOCK
     if (! findWindowTypeProperty(WINDOW_TYPE_DOCK)) {
         AutoProperty *prop = new AutoProperty();
         prop->maskAdd(AP_STICKY);
         prop->sticky = true;
         prop->maskAdd(AP_TITLEBAR);
-        prop->border = false;
+        prop->titlebar = false;
         prop->maskAdd(AP_BORDER);
         prop->border = false;
         prop->maskAdd(AP_SKIP);
         prop->skip = SKIP_MENUS|SKIP_FOCUS_TOGGLE|SKIP_PAGER|SKIP_TASKBAR;
         prop->maskAdd(AP_LAYER);
         prop->layer = LAYER_DOCK;
+        prop->maskAdd(AP_FOCUSABLE);
+        prop->focusable = false;
+        prop->maskAdd(AP_DISALLOWED_ACTIONS);
+        prop->disallowed_actions = ACTION_ACCESS_MOVE|ACTION_ACCESS_RESIZE;
 
         _window_type_prop_map[WINDOW_TYPE_DOCK] = prop;
     }
@@ -594,9 +650,9 @@ AutoProperties::setDefaultTypeProperties(void)
     if (! findWindowTypeProperty(WINDOW_TYPE_TOOLBAR)) {
         AutoProperty *prop = new AutoProperty();
         prop->maskAdd(AP_TITLEBAR);
-        prop->border = false;
+        prop->titlebar = true;
         prop->maskAdd(AP_BORDER);
-        prop->border = false;
+        prop->border = true;
         prop->maskAdd(AP_SKIP);
         prop->skip = SKIP_MENUS|SKIP_FOCUS_TOGGLE|SKIP_PAGER|SKIP_TASKBAR;
 
@@ -607,7 +663,7 @@ AutoProperties::setDefaultTypeProperties(void)
     if (! findWindowTypeProperty(WINDOW_TYPE_MENU)) {
         AutoProperty *prop = new AutoProperty();
         prop->maskAdd(AP_TITLEBAR);
-        prop->border = false;
+        prop->titlebar = false;
         prop->maskAdd(AP_BORDER);
         prop->border = false;
         prop->maskAdd(AP_SKIP);
@@ -620,9 +676,9 @@ AutoProperties::setDefaultTypeProperties(void)
     if (! findWindowTypeProperty(WINDOW_TYPE_UTILITY)) {
         AutoProperty *prop = new AutoProperty();
         prop->maskAdd(AP_TITLEBAR);
-        prop->border = false;
+        prop->titlebar = true;
         prop->maskAdd(AP_BORDER);
-        prop->border = false;
+        prop->border = true;
         prop->maskAdd(AP_SKIP);
         prop->skip = SKIP_MENUS|SKIP_FOCUS_TOGGLE|SKIP_SNAP;
 
@@ -633,7 +689,7 @@ AutoProperties::setDefaultTypeProperties(void)
     if (! findWindowTypeProperty(WINDOW_TYPE_SPLASH)) {
         AutoProperty *prop = new AutoProperty();
         prop->maskAdd(AP_TITLEBAR);
-        prop->border = false;
+        prop->titlebar = false;
         prop->maskAdd(AP_BORDER);
         prop->border = false;
 
@@ -712,9 +768,11 @@ AutoProperties::parseAutoPropertyValue(CfgParser::Entry *section, AutoProperty *
                        &prop->client_gm.x, &prop->client_gm.y,
                        &prop->client_gm.width, &prop->client_gm.height);
             break;
-        case AP_LAYER:
-            prop->maskAdd(AP_LAYER);
+        case AP_LAYER:            
             prop->layer = Config::instance()->getLayer((*it)->get_value());
+            if (prop->layer != LAYER_NONE) {
+                prop->maskAdd(AP_LAYER);
+            }
             break;
         case AP_WORKSPACE:
             prop->maskAdd(AP_WORKSPACE);
@@ -754,7 +812,20 @@ AutoProperties::parseAutoPropertyValue(CfgParser::Entry *section, AutoProperty *
                 }
             }
             break;
-
+        case AP_ALLOWED_ACTIONS:
+            prop->maskAdd(AP_ALLOWED_ACTIONS);
+            Config::instance()->parseActionAccessMask((*it)->get_value(), prop->allowed_actions);
+            break;
+        case AP_DISALLOWED_ACTIONS:
+            prop->maskAdd(AP_DISALLOWED_ACTIONS);
+            Config::instance()->parseActionAccessMask((*it)->get_value(), prop->disallowed_actions);
+            break;
+#ifdef OPACITY
+        case AP_OPACITY:
+            prop->maskAdd(AP_OPACITY);
+            Config::parseOpacity((*it)->get_value(), prop->focus_opacity, prop->unfocus_opacity);
+            break;
+#endif // OPACITY
         default:
             // do nothing
             break;
@@ -783,22 +854,20 @@ AutoProperties::findDecorProperty(const ClassHint* class_hint)
     return static_cast<DecorProperty*>(findProperty(class_hint, &_decor_prop_list, -1, 0));
 }
 
-#ifdef HARBOUR
 DockAppProperty*
 AutoProperties::findDockAppProperty(const ClassHint *class_hint)
 {
     return static_cast<DockAppProperty*>(findProperty(class_hint, &_dock_app_prop_list, -1, 0));
 }
-#endif // HARBOUR
 
 //! @brief Get AutoProperty for window of type type
 //! @param atom Atom to get property for.
 //! @return AutoProperty on success, else 0.
 AutoProperty*
-AutoProperties::findWindowTypeProperty(EwmhAtomName atom)
+AutoProperties::findWindowTypeProperty(AtomName atom)
 {
     AutoProperty *prop = 0;
-    map<EwmhAtomName, AutoProperty*>::iterator it(_window_type_prop_map.find(atom));
+    map<AtomName, AutoProperty*>::iterator it(_window_type_prop_map.find(atom));
     if (it != _window_type_prop_map.end()) {
         prop = it->second;
     }
