@@ -1,13 +1,12 @@
 //
 // FrameListMenu.cc for pekwm
-// Copyright (C) 2002-2004 Claes Nasten <pekdon{@}pekdon{.}net>
+// Copyright (C) 2002-2009 Claes Nasten <pekdon{@}pekdon{.}net>
 //
 // This program is licensed under the GNU GPL.
 // See the LICENSE file for more information.
 //
 
 #include "../config.h"
-#ifdef MENUS
 
 #include "PWinObj.hh"
 #include "PDecor.hh"
@@ -51,7 +50,7 @@ DecorMenu::handleItemExec(PMenu::Item *item)
         return;
     }
 
-    ActionPerformed ap(_wo_ref, item->getAE());
+    ActionPerformed ap(getWORef(), item->getAE());
     _act->handleAction(ap);
 }
 
@@ -78,5 +77,3 @@ DecorMenu::reload(CfgParser::Entry *section)
 
     buildMenu(); // rebuild the menu
 }
-
-#endif // MENUS
