@@ -1,6 +1,6 @@
 //
 // Frame.cc for pekwm
-// Copyright © 2002-2008 Claes Nästén <me@pekdon.net>
+// Copyright © 2002-2009 Claes Nästén <me@pekdon.net>
 //
 // This program is licensed under the GNU GPL.
 // See the LICENSE file for more information.
@@ -12,6 +12,7 @@
 
 #include <algorithm>
 #include <cstdio>
+#include <cwchar>
 #include <functional>
 #include <iostream>
 
@@ -34,14 +35,7 @@ extern "C" {
 #include "Workspaces.hh"
 #include "WindowManager.hh"
 #include "KeyGrabber.hh"
-#ifdef HARBOUR
-#include "Harbour.hh"
-#endif // HARBOUR
-
-#ifdef MENUS
 #include "PMenu.hh"
-#endif //MENUS
-
 #include "Theme.hh"
 
 using std::cerr;
@@ -52,6 +46,7 @@ using std::mem_fun;
 using std::string;
 using std::vector;
 using std::wstring;
+using std::swprintf;
 
 list<Frame*> Frame::_frame_list = list<Frame*>();
 vector<uint> Frame::_frameid_list = vector<uint>();
@@ -61,13 +56,13 @@ bool Frame::_tag_behind = false;
 
 //! @brief Frame constructor
 Frame::Frame(Client *client, AutoProperty *ap)
-    : PDecor(WindowManager::inst()->getScreen()->getDpy(), WindowManager::inst()->getTheme(),
-             Frame::getClientDecorName(client)),
+    : PDecor(WindowManager::instance()->getTheme(),
+             Frame::getClientDecorName(client), client->getWindow()),
       _id(0), _client(0), _class_hint(0),
       _non_fullscreen_decor_state(0), _non_fullscreen_layer(LAYER_NORMAL)
 {
     // setup basic pointers
-    _scr = WindowManager::inst()->getScreen();
+    _scr = WindowManager::instance()->getScreen();
 
     // PWinObj attributes
     _type = WO_FRAME;
@@ -86,11 +81,9 @@ Frame::Frame(Client *client, AutoProperty *ap)
     }
 
     // get unique id of the frame, if the client didn't have an id
-    if (! WindowManager::inst()->isStartup()) {
+    if (! WindowManager::instance()->isStartup()) {
         long id;
-
-        if (AtomUtil::getLong(client->getWindow(),
-                              PekwmAtoms::instance()->getAtom(PEKWM_FRAME_ID), id)) {
+        if (AtomUtil::getLong(client->getWindow(), Atoms::getAtom(PEKWM_FRAME_ID), id)) {
             _id = id;
         }
 
@@ -126,20 +119,22 @@ Frame::Frame(Client *client, AutoProperty *ap)
     if (client->isViewable() || client->isPlaced()) {
         moveChild(client->getX(), client->getY());
     } else if (client->setPUPosition()) {
-        calcGravityPosition(client->getXSizeHints()->win_gravity,
-                            client->getX(), client->getY(), _gm.x, _gm.y);
-        move(_gm.x, _gm.y);
+        int x, y;
+        calcGravityPosition(client->getXSizeHints()->win_gravity, client->getX(), client->getY(), x, y);
+        move(x, y);
     } else {
         place = Config::instance()->isPlaceNew();
     }
 
     // override both position and size with autoproperties
+    bool ap_geometry = false;
     if (ap) {
         if (ap->isMask(AP_FRAME_GEOMETRY|AP_CLIENT_GEOMETRY)) {
+            ap_geometry = true;
+
             setupAPGeometry(client, ap);
 
-            if (ap->frame_gm_mask&(XValue|YValue) ||
-                    ap->client_gm_mask&(XValue|YValue)) {
+            if (ap->frame_gm_mask&(XValue|YValue) || ap->client_gm_mask&(XValue|YValue)) {
                 place = false;
             }
         }
@@ -151,7 +146,7 @@ Frame::Frame(Client *client, AutoProperty *ap)
 
     // still need a position?
     if (place) {
-        Workspaces::instance()->placeWo(this, client->getTransientWindow());
+        Workspaces::instance()->placeWo(this, client->getTransientClientWindow());
     }
 
     _old_gm = _gm;
@@ -160,23 +155,25 @@ Frame::Frame(Client *client, AutoProperty *ap)
 
     _scr->ungrabServer(true); // ungrab and sync
 
-    // needs to be done before the workspace insert
+    // now insert the client in the frame we created.
+    addChild(client);
+
+    // needs to be done before the workspace insert and after the client
+    // has been inserted, in order for layer settings to be propagated
     setLayer(client->getLayer());
 
     // I add these to the list before I insert the client into the frame to
     // be able to skip an extra updateClientList
     _frame_list.push_back(this);
-    WindowManager::inst()->addToFrameList(this);
+    WindowManager::instance()->addToFrameList(this);
     Workspaces::instance()->insert(this);
 
-    // now insert the client in the frame we created, do not give it focus.
-    addChild(client);
     activateChild(client);
 
     // set the window states, shaded, maximized...
     getState(client);
 
-    if (! _client->hasStrut()) {
+    if (! _client->hasStrut() && ! ap_geometry) {
         if (fixGeometry()) {
             moveResize(_gm.x, _gm.y, _gm.width, _gm.height);
         }
@@ -200,7 +197,7 @@ Frame::~Frame(void)
     woListRemove(this);
     _frame_list.remove(this);
     Workspaces::instance()->remove(this);
-    WindowManager::inst()->removeFromFrameList(this);
+    WindowManager::instance()->removeFromFrameList(this);
     if (_tag_frame == this) {
         _tag_frame = 0;
     }
@@ -211,7 +208,8 @@ Frame::~Frame(void)
         delete _class_hint;
     }
 
-    Workspaces::instance()->updateClientStackingList(true, true);
+    Workspaces::instance()->updateClientList();
+    Workspaces::instance()->updateClientStackingList();
 }
 
 // START - PWinObj interface.
@@ -241,9 +239,26 @@ Frame::stick(void)
     PDecor::setWorkspace(Workspaces::instance()->getActive());
 }
 
+void
+Frame::raise(void)
+{
+    PDecor::raise();
+
+    if (_client->transient_size()) {
+        Frame *frame;
+        list<Client*>::iterator it(_client->transient_begin());
+        for (; it != _client->transient_end(); ++it) {
+            frame = static_cast<Frame*>((*it)->getParent());
+            if (frame != this && frame->getActiveClient() == *it) {
+                frame->raise();
+            }
+        }
+    }
+}
+
 //! @brief Sets workspace on frame, wrapper to allow autoproperty loading
 void
-Frame::setWorkspace(uint workspace)
+Frame::setWorkspace(unsigned int workspace)
 {
     // Duplicate the behavior done in PDecor::setWorkspace to have a sane
     // value on _workspace and not NET_WM_STICKY_WINDOW.
@@ -258,9 +273,20 @@ Frame::setWorkspace(uint workspace)
     PDecor::setWorkspace(workspace);
 }
 
+void
+Frame::setLayer(unsigned int layer)
+{
+    PDecor::setLayer(layer);
+
+    if (_client) {
+        LayerObservation *observation = new LayerObservation(Layer(layer));
+        _client->notifyObservers(observation);
+        delete observation;
+    }
+}
+
 // event handlers
 
-//! @brief
 ActionEvent*
 Frame::handleMotionEvent(XMotionEvent *ev)
 {
@@ -356,7 +382,6 @@ Frame::handleLeaveEvent(XCrossingEvent *ev)
     return ae;
 }
 
-//! @brief
 ActionEvent*
 Frame::handleMapRequest(XMapRequestEvent *ev)
 {
@@ -377,7 +402,6 @@ Frame::handleMapRequest(XMapRequestEvent *ev)
     return 0;
 }
 
-//! @brief
 ActionEvent*
 Frame::handleUnmapEvent(XUnmapEvent *ev)
 {
@@ -394,35 +418,71 @@ Frame::handleUnmapEvent(XUnmapEvent *ev)
 
 // END - PWinObj interface.
 
-#ifdef HAVE_SHAPE
-//! @brief
 void
 Frame::handleShapeEvent(XAnyEvent *ev)
 {
+#ifdef HAVE_SHAPE
     if (! _client || (ev->window != _client->getWindow())) {
         return;
     }
     _client->setShaped(setShape());
-}
 #endif // HAVE_SHAPE
+}
 
 // START - PDecor interface.
 
+bool
+Frame::allowMove(void) const
+{
+    return _client ? _client->allowMove() : true;
+}
+
+/**
+ * Return active client, or 0 if no clients or active child is not a Client.
+ */
+Client*
+Frame::getActiveClient(void)
+{
+    if (getActiveChild() && getActiveChild()->getType() == WO_CLIENT) {
+        return static_cast<Client*>(getActiveChild());
+    } else {
+        return 0;
+    }
+}
+
 //! @brief Adds child to the frame.
 void
-Frame::addChild (PWinObj *child)
+Frame::addChild (PWinObj *child, std::list<PWinObj*>::iterator *it)
 {
-    PDecor::addChild (child);
-    AtomUtil::setLong (child->getWindow (),
-                       PekwmAtoms::instance ()->getAtom (PEKWM_FRAME_ID), _id);
-    child->lower ();
+    PDecor::addChild(child, it);
+    AtomUtil::setLong(child->getWindow(), Atoms::getAtom(PEKWM_FRAME_ID), _id);
+    child->lower();
+}
+
+/**
+ * Add child preserving order from the previous pekwm run.
+ */
+void
+Frame::addChildOrdered (Client *child)
+{
+    Client *client;
+    
+    list<PWinObj*>::iterator it(_child_list.begin());
+    for (; it != _child_list.end(); ++it) {
+        client = static_cast<Client*>(*it);
+        if (child->getInitialFrameOrder() < client->getInitialFrameOrder()) {
+            break;
+        }
+    }
+
+    addChild(child, &it);
 }
 
 //! @brief Removes child from the frame.
 void
 Frame::removeChild (PWinObj *child, bool do_delete)
 {
-    PDecor::removeChild (child, do_delete);
+    PDecor::removeChild(child, do_delete);
 }
 
 /**
@@ -432,6 +492,9 @@ Frame::removeChild (PWinObj *child, bool do_delete)
 void
 Frame::activateChild (PWinObj *child)
 {
+    // FIXME: Update default decoration for this child, can change
+    // decoration from DEFAULT to REMOTE or WARNING
+
     // Sync the frame state with the client only if we already had a client
     if (_client && _client != child) {
         applyState(static_cast<Client*>(child));
@@ -442,11 +505,13 @@ Frame::activateChild (PWinObj *child)
 
     // setShape uses current active child, so we need to activate the
     // child before setting shape
-#ifdef HAVE_SHAPE
     if (PScreen::instance()->hasExtensionShape()) {
         _client->setShaped(setShape());
     }
-#endif // HAVE_SHAPE
+
+#ifdef OPACITY
+    setOpacity(_client);
+#endif // OPACITY
 
     if (_focused) {
         child->giveInputFocus();
@@ -455,24 +520,33 @@ Frame::activateChild (PWinObj *child)
     // Reload decor rules if needed.
     handleTitleChange(_client);
 
-    Workspaces::instance()->updateClientStackingList(true, true);
+    Workspaces::instance()->updateClientList();
+    Workspaces::instance()->updateClientStackingList();
 }
 
-//! @brief
+/**
+ * Called when child order is updated, re-sets the titles and updates
+ * the _PEKWM_FRAME hints.
+ */
 void
 Frame::updatedChildOrder(void)
 {
     titleClear();
 
+    Client *client;
     list<PWinObj*>::iterator it(_child_list.begin());
-    for (; it != _child_list.end(); ++it) {
-        titleAdd(static_cast<Client*>(*it)->getTitle());
+    for (long num = 0; it != _child_list.end(); ++num, ++it) {
+        client = static_cast<Client*>(*it);
+        client->setPekwmFrameOrder(num);
+        titleAdd(client->getTitle());
     }
 
     updatedActiveChild();
 }
 
-//! @brief
+/**
+ * Set the active title and update the _PEKWM_FRAME hints.
+ */
 void
 Frame::updatedActiveChild(void)
 {
@@ -480,9 +554,9 @@ Frame::updatedActiveChild(void)
 
     list<PWinObj*>::iterator it(_child_list.begin());
     for (uint i = 0; it != _child_list.end(); ++i, ++it) {
+        static_cast<Client*>(*it)->setPekwmFrameActive(_child == *it);
         if (_child == *it) {
             titleSetActive(i);
-            break;
         }
     }
 
@@ -604,7 +678,7 @@ Frame::setId(uint id)
     _id = id;
 
     list<PWinObj*>::iterator it(_child_list.begin());
-    long atom = PekwmAtoms::instance()->getAtom(PEKWM_FRAME_ID);
+    long atom = Atoms::getAtom(PEKWM_FRAME_ID);
     for (; it != _child_list.end(); ++it) {
         AtomUtil::setLong((*it)->getWindow(), atom, id);
     }
@@ -627,8 +701,9 @@ Frame::getState(Client *cl)
         setStateMaximized(STATE_TOGGLE, false, true, false);
     if (isShaded() != cl->isShaded())
         setShaded(STATE_TOGGLE);
-    if (_layer != cl->getLayer())
+    if (getLayer() != cl->getLayer()) {
         setLayer(cl->getLayer());
+    }
     if (_workspace != cl->getWorkspace())
         PDecor::setWorkspace(cl->getWorkspace());
 
@@ -667,7 +742,7 @@ Frame::applyState(Client *cl)
     cl->setMaximizedVert(_maximized_vert);
     cl->setShade(isShaded());
     cl->setWorkspace(_workspace);
-    cl->setLayer(_layer);
+    cl->setLayer(getLayer());
 
     // fix border / titlebar state
     cl->setBorder(hasBorder());
@@ -840,7 +915,7 @@ Frame::returnFrameID(uint id)
 std::string
 Frame::getClientDecorName(Client *client)
 {
-  DecorProperty *prop = WindowManager::inst()->getAutoProperties()->findDecorProperty(client->getClassHint());
+  DecorProperty *prop = WindowManager::instance()->getAutoProperties()->findDecorProperty(client->getClassHint());
   return prop ? prop->getName() : PDecor::DEFAULT_DECOR_NAME;
 }
 
@@ -960,7 +1035,7 @@ Frame::doGroupingDrag(XMotionEvent *ev, Client *client, bool behind) // FIXME: r
             Client *search = 0;
 
             // only group if we have grouping turned on
-            if (WindowManager::inst()->isAllowGrouping()) {
+            if (WindowManager::instance()->isAllowGrouping()) {
                 int x, y;
                 Window win;
 
@@ -1094,6 +1169,7 @@ Frame::doResize(bool left, bool x, bool top, bool y)
     }
 
     setShaded(STATE_UNSET); // make sure the frame isn't shaded
+    setStateFullscreen(STATE_UNSET);
 
     // Initialize variables
     int start_x, new_x;
@@ -1116,11 +1192,12 @@ Frame::doResize(bool left, bool x, bool top, bool y)
     wchar_t buf[128];
     getDecorInfo(buf, 128);
 
+    bool center_on_root = Config::instance()->isShowStatusWindowOnRoot();
     StatusWindow *sw = StatusWindow::instance();
     if (Config::instance()->isShowStatusWindow()) {
-        sw->draw(buf, true);
+        sw->draw(buf, true, center_on_root ? 0 : &_gm);
         sw->mapWindowRaised();
-        sw->draw(buf);
+        sw->draw(buf, true, center_on_root ? 0 : &_gm);
     }
 
     bool outline = ! Config::instance()->getOpaqueResize();
@@ -1130,19 +1207,23 @@ Frame::doResize(bool left, bool x, bool top, bool y)
         _scr->grabServer();
     }
 
+    const long resize_mask = ButtonPressMask|ButtonReleaseMask|ButtonMotionMask;
     XEvent ev;
     bool exit = false;
     while (exit != true) {
         if (outline) {
             drawOutline(_gm);
         }
-        XMaskEvent(_dpy, ButtonPressMask|ButtonReleaseMask|ButtonMotionMask, &ev);
+        XMaskEvent(_dpy, resize_mask, &ev);
         if (outline) {
             drawOutline(_gm); // clear
         }
 
         switch (ev.type) {
         case MotionNotify:
+            // Flush all pointer motion, no need to redraw and redraw.
+            PScreen::removeMotionEvents();
+
             if (x) {
                 new_x = start_x - pointer_x + ev.xmotion.x;
             }
@@ -1154,7 +1235,7 @@ Frame::doResize(bool left, bool x, bool top, bool y)
 
             getDecorInfo(buf, 128);
             if (Config::instance()->isShowStatusWindow()) {
-                sw->draw(buf, true);
+                sw->draw(buf, true, center_on_root ? 0 : &_gm);
             }
 
             // only updated when needed when in opaque mode
@@ -1222,7 +1303,8 @@ Frame::recalcResizeDrag(int nx, int ny, bool left, bool top)
 
     width -= brdr_lr;
     height -= brdr_tb + getTitleHeight();
-    _client->getIncSize(&width, &height, width, height);
+
+    _client->getAspectSize(&width, &height, width, height);
 
     const XSizeHints *hints = _client->getXSizeHints();
     // check so we aren't overriding min or max size
@@ -1342,11 +1424,12 @@ void
 Frame::setStateMaximized(StateAction sa, bool horz, bool vert, bool fill)
 {
     // we don't want to maximize transients
-    if (_client->getTransientWindow()) {
+    if (_client->isTransient()) {
         return;
     }
 
     setShaded(STATE_UNSET);
+    setStateFullscreen(STATE_UNSET);
 
     // make sure the two states are in sync if toggling
     if ((horz == vert) && (sa == STATE_TOGGLE)) {
@@ -1435,7 +1518,7 @@ Frame::setStateMaximized(StateAction sa, bool horz, bool vert, bool fill)
     }
 
     fixGeometry(); // harbour already considered
-    downSize(true, true); // keep x and keep y ( make conform to inc size )
+    downSize(_gm, true, true); // keep x and keep y ( make conform to inc size )
 
     moveResize(_gm.x, _gm.y, _gm.width, _gm.height);
 
@@ -1483,7 +1566,7 @@ Frame::setStateFullscreen(StateAction sa)
 
     // Re-stack window if fullscreen is above other windows.
     if (Config::instance()->isFullscreenAbove()) {
-        _layer = _fullscreen ? LAYER_ABOVE_DOCK : _non_fullscreen_layer;
+        setLayer(_fullscreen ? LAYER_ABOVE_DOCK : _non_fullscreen_layer);
         raise();
     }
     
@@ -1502,29 +1585,27 @@ Frame::setStateSticky(StateAction sa)
     }
 }
 
-//! @brief
 void
 Frame::setStateAlwaysOnTop(StateAction sa)
 {
-    if (! ActionUtil::needToggle(sa, _layer == LAYER_ONTOP)) {
+    if (! ActionUtil::needToggle(sa, getLayer() == LAYER_ONTOP)) {
         return;
     }
 
-    _client->alwaysOnTop(_layer < LAYER_ONTOP);
+    _client->alwaysOnTop(getLayer() < LAYER_ONTOP);
     setLayer(_client->getLayer());
 
     raise();
 }
 
-//! @brief
 void
 Frame::setStateAlwaysBelow(StateAction sa)
 {
-    if (! ActionUtil::needToggle(sa, _layer == LAYER_BELOW)) {
+    if (! ActionUtil::needToggle(sa, getLayer() == LAYER_BELOW)) {
         return;
     }
 
-    _client->alwaysBelow(_layer > LAYER_BELOW);
+    _client->alwaysBelow(getLayer() > LAYER_BELOW);
     setLayer(_client->getLayer());
 
     lower();
@@ -1545,7 +1626,7 @@ Frame::setStateDecorBorder(StateAction sa)
 
         // update the _PEKWM_FRAME_DECOR hint
         AtomUtil::setLong(_client->getWindow(),
-                          PekwmAtoms::instance()->getAtom(PEKWM_FRAME_DECOR),
+                          Atoms::getAtom(PEKWM_FRAME_DECOR),
                           _client->getDecorState());
     }
 }
@@ -1564,7 +1645,7 @@ Frame::setStateDecorTitlebar(StateAction sa)
         _client->setTitlebar(hasTitlebar());
 
         AtomUtil::setLong(_client->getWindow(),
-                          PekwmAtoms::instance()->getAtom(PEKWM_FRAME_DECOR),
+                          Atoms::getAtom(PEKWM_FRAME_DECOR),
                           _client->getDecorState());
     }
 }
@@ -1623,7 +1704,7 @@ Frame::setStateTitle(StateAction sa, Client *client, const std::wstring &title)
 
     } else if (sa == STATE_UNSET) {
         client->getTitle()->setUser(L"");
-        client->getXClientName();
+        client->readName();
     } else {
         if (client->getTitle()->isUserSet()) {
             client->getTitle()->setUser(L"");
@@ -1634,9 +1715,9 @@ Frame::setStateTitle(StateAction sa, Client *client, const std::wstring &title)
 
     // Set PEKWM_TITLE atom to preserve title on client between sessions.
     AtomUtil::setString(client->getWindow(),
-                        PekwmAtoms::instance()->getAtom(PEKWM_TITLE),
+                        Atoms::getAtom(PEKWM_TITLE),
                         Util::to_mb_str(client->getTitle()->getUser()));
-    
+
 
     renderTitle();
 }
@@ -1656,6 +1737,17 @@ Frame::setStateMarked(StateAction sa, Client *client)
     renderTitle();
 }
 
+#ifdef OPACITY
+void
+Frame::setStateOpaque(StateAction sa)
+{
+    if (! ActionUtil::needToggle(sa, _opaque)) {
+        return;
+    }
+    _client->setOpaque(!_opaque);
+    setOpaque(!_opaque);
+}
+#endif // OPACITY
 // STATE actions end
 
 //! @brief
@@ -1724,7 +1816,7 @@ Frame::growDirection(uint direction)
         break;
     }
 
-    downSize((direction != DIRECTION_LEFT), (direction != DIRECTION_UP));
+    downSize(_gm, (direction != DIRECTION_LEFT), (direction != DIRECTION_UP));
 
     moveResize(_gm.x, _gm.y, _gm.width, _gm.height);
 }
@@ -1759,9 +1851,9 @@ Frame::readAutoprops(uint type)
     // Set the correct group of the window
     _class_hint->group = data->group_name;
 
-    if (_class_hint == _client->getClassHint() &&
-            (_client->getTransientWindow() &&
-             ! data->isApplyOn(APPLY_ON_TRANSIENT))) {
+    if (_class_hint == _client->getClassHint()
+        && (_client->isTransient()
+            && ! data->isApplyOn(APPLY_ON_TRANSIENT))) {
         return;
     }
 
@@ -1771,12 +1863,12 @@ Frame::readAutoprops(uint type)
     if (data->isMask(AP_SHADED) && (isShaded() != data->shaded)) {
         setShaded(STATE_UNSET);
     }
-    if (data->isMask(AP_MAXIMIZED_HORIZONTAL) &&
-            (_maximized_horz != data->maximized_horizontal)) {
+    if (data->isMask(AP_MAXIMIZED_HORIZONTAL)
+        && (_maximized_horz != data->maximized_horizontal)) {
         setStateMaximized(STATE_TOGGLE, true, false, false);
     }
-    if (data->isMask(AP_MAXIMIZED_VERTICAL) &&
-            (_maximized_vert != data->maximized_vertical)) {
+    if (data->isMask(AP_MAXIMIZED_VERTICAL)
+        && (_maximized_vert != data->maximized_vertical)) {
         setStateMaximized(STATE_TOGGLE, false, true, false);
     }
     if (data->isMask(AP_FULLSCREEN) && (_fullscreen != data->fullscreen)) {
@@ -1899,9 +1991,11 @@ Frame::calcGravityPosition(int gravity, int x, int y, int &g_x, int &g_y)
     }
 }
 
-//! @brief Makes Frame conform to Clients width and height inc
+/**
+ * Makes gm conform to _clients width and height inc.
+ */
 void
-Frame::downSize(bool keep_x, bool keep_y)
+Frame::downSize(Geometry &gm, bool keep_x, bool keep_y)
 {
     XSizeHints *size_hint = _client->getXSizeHints(); // convenience
 
@@ -1912,9 +2006,9 @@ Frame::downSize(bool keep_x, bool keep_y)
                   ? size_hint->base_width
                   : (size_hint->flags&PMinSize) ? size_hint->min_width : 0;
 
-        _gm.width -= (getChildWidth() - b_x) % size_hint->width_inc;
+        gm.width -= (getChildWidth() - b_x) % size_hint->width_inc;
         if (! keep_x) {
-            _gm.x = o_r - _gm.width;
+            gm.x = o_r - gm.width;
         }
     }
 
@@ -1925,9 +2019,9 @@ Frame::downSize(bool keep_x, bool keep_y)
                   ? size_hint->base_height
                   : (size_hint->flags&PMinSize) ? size_hint->min_height : 0;
 
-        _gm.height -= (getChildHeight() - b_y) % size_hint->height_inc;
+        gm.height -= (getChildHeight() - b_y) % size_hint->height_inc;
         if (! keep_y) {
-            _gm.y = o_b - _gm.height;
+            gm.y = o_b - gm.height;
         }
     }
 }
@@ -1989,119 +2083,73 @@ Frame::handleConfigureRequest(XConfigureRequestEvent *ev, Client *client)
 void
 Frame::handleConfigureRequestGeometry(XConfigureRequestEvent *ev, Client *client)
 {
-    // Look for fullscreen requests
-    long all_geometry = CWX|CWY|CWWidth|CWHeight;
-    bool is_fullscreen = false;
-    if (Config::instance()->isFullscreenDetect()
-        && ! client->isCfgDeny(CFG_DENY_SIZE)
-        && ! client->isCfgDeny(CFG_DENY_POSITION)
-        && ((ev->value_mask&all_geometry) == static_cast<unsigned>(all_geometry))) {
-
-        Geometry gm_request(ev->x, ev->y, ev->width, ev->height);
-        if (gm_request == _scr->getScreenGeometry()
-             || gm_request == _scr->getHeadGeometry(_scr->getNearestHead(ev->x, ev->y))) {
-
-            is_fullscreen = true;
-            setStateFullscreen(STATE_SET);
-        }
+    if (isRequestGeometryFullscreen(ev, client)) {
+        client->configureRequestSend();
+        return;
     }
 
-    if (! is_fullscreen) {
-        // Remove fullscreen state if client changes it size
-        if (Config::instance()->isFullscreenDetect()) {
-            setStateFullscreen(STATE_UNSET);
-        }
+    bool change_geometry = false;
+    if (! client->isCfgDeny(CFG_DENY_SIZE)
+        && (ev->value_mask & (CWWidth|CWHeight))) {
+        resizeChild(ev->width, ev->height);
+        _client->setShaped(setShape());
+        change_geometry = true;
+    }
 
-        if (! client->isCfgDeny(CFG_DENY_SIZE)
-                && (ev->value_mask & (CWWidth|CWHeight)) ) {
+    if (! client->isCfgDeny(CFG_DENY_POSITION)
+        && (ev->value_mask & (CWX|CWY)) ) {
+        calcGravityPosition(_client->getXSizeHints()->win_gravity,
+                            ev->x, ev->y, _gm.x, _gm.y);
+        move(_gm.x, _gm.y);
+        change_geometry = true;
+    }
 
-            resizeChild(ev->width, ev->height);
-            _client->setShaped(setShape());
-        }
-
-        if (! client->isCfgDeny(CFG_DENY_POSITION)
-               && (ev->value_mask & (CWX|CWY)) ) {
-
-            calcGravityPosition(_client->getXSizeHints()->win_gravity,
-                                ev->x, ev->y, _gm.x, _gm.y);
-            move(_gm.x, _gm.y);
-        }
+    // Remove fullscreen state if client changes it size
+    if (change_geometry && Config::instance()->isFullscreenDetect()) {
+        setStateFullscreen(STATE_UNSET);
     }
 }
 
-//! @brief
+/**
+ * Check if requested size if "fullscreen"
+ */
+bool
+Frame::isRequestGeometryFullscreen(XConfigureRequestEvent *ev, Client *client)
+{
+    bool is_fullscreen = false;
+    if (! client->isCfgDeny(CFG_DENY_SIZE)
+        && ! client->isCfgDeny(CFG_DENY_POSITION)) {
+        int nearest_head = _scr->getNearestHead(ev->x, ev->y);
+        Geometry gm_request(ev->x, ev->y, ev->width, ev->height);
+        Geometry gm_screen(_scr->getScreenGeometry());
+        Geometry gm_head(_scr->getHeadGeometry(nearest_head));
+
+        if (gm_request == gm_screen || gm_request == gm_head) {
+            is_fullscreen = true;
+        } else {
+            downSize(gm_screen, true, true);
+            downSize(gm_head, true, true);
+            if (gm_request == gm_screen || gm_request == gm_head) {
+                is_fullscreen = true;
+            }
+        }
+
+        if (is_fullscreen && Config::instance()->isFullscreenDetect()) {
+            setStateFullscreen(STATE_SET);
+        }
+    }
+    return is_fullscreen;
+}
+
+/**
+ * Handle client message.
+ */
 void
 Frame::handleClientMessage(XClientMessageEvent *ev, Client *client)
 {
-    EwmhAtoms *ewmh = EwmhAtoms::instance(); // convenience
-
-    StateAction sa;
-
-    if (ev->message_type == ewmh->getAtom(STATE)) {
-        if (ev->data.l[0]== NET_WM_STATE_REMOVE) {
-            sa = STATE_UNSET;
-        } else if (ev->data.l[0]== NET_WM_STATE_ADD) {
-            sa = STATE_SET;
-        } else if (ev->data.l[0]== NET_WM_STATE_TOGGLE) {
-            sa = STATE_TOGGLE;
-        } else {
-#ifdef DEBUG
-            cerr << __FILE__ << "@" << __LINE__ << ": "
-                 << "None of StateAdd, StateRemove or StateToggle." << endl;
-#endif // DEBUG
-            return;
-        }
-
-#define IS_STATE(S) ((ev->data.l[1] == long(ewmh->getAtom(S))) || (ev->data.l[2] == long(ewmh->getAtom(S))))
-
-        // actions that only is going to be applied on the active client
-        if (client == _client) {
-            // there is no modal support in pekwm yet
-// 			if (IS_STATE(STATE_MODAL)) {
-// 				is_modal=true;
-// 			}
-            if (IS_STATE(STATE_STICKY)) {
-                setStateSticky(sa);
-            }
-            if (IS_STATE(STATE_MAXIMIZED_HORZ)
-                    && ! client->isCfgDeny(CFG_DENY_STATE_MAXIMIZED_HORZ)) {
-                setStateMaximized(sa, true, false, false);
-            }
-            if (IS_STATE(STATE_MAXIMIZED_VERT)
-                    && ! client->isCfgDeny(CFG_DENY_STATE_MAXIMIZED_VERT)) {
-                setStateMaximized(sa, false, true, false);
-            }
-            if (IS_STATE(STATE_SHADED)) {
-                setShaded(sa);
-            }
-            if (IS_STATE(STATE_HIDDEN)
-                    && ! client->isCfgDeny(CFG_DENY_STATE_HIDDEN)) {
-                setStateIconified(sa);
-            }
-            if (IS_STATE(STATE_FULLSCREEN)
-                    && ! client->isCfgDeny(CFG_DENY_STATE_FULLSCREEN)) {
-                setStateFullscreen(sa);
-            }
-            if (IS_STATE(STATE_ABOVE)
-                    && ! client->isCfgDeny(CFG_DENY_STATE_ABOVE)) {
-                setStateAlwaysOnTop(sa);
-            }
-            if (IS_STATE(STATE_BELOW)
-                    && ! client->isCfgDeny(CFG_DENY_STATE_BELOW)) {
-                setStateAlwaysBelow(sa);
-            }
-        }
-
-        if (IS_STATE(STATE_SKIP_TASKBAR)) {
-            client->setStateSkip(sa, SKIP_TASKBAR);
-        }
-        if (IS_STATE(STATE_SKIP_PAGER)) {
-            client->setStateSkip(sa, SKIP_PAGER);
-        }
-
-        client->updateEwmhStates();
-
-    } else if (ev->message_type == ewmh->getAtom(NET_ACTIVE_WINDOW)) {
+    if (ev->message_type == Atoms::getAtom(STATE)) {
+        handleClientStateMessage(ev, client);
+    } else if (ev->message_type == Atoms::getAtom(NET_ACTIVE_WINDOW)) {
         if (! client->isCfgDeny(CFG_DENY_ACTIVE_WINDOW)) {
             // Active child if it's not the active child
             if (client != _client) {
@@ -2111,7 +2159,8 @@ Frame::handleClientMessage(XClientMessageEvent *ev, Client *client)
             // If we aren't mapped we check if we make sure we're on the right
             // workspace and then map the window.
             if (! _mapped) {
-                if (_workspace != Workspaces::instance()->getActive()) {
+                if (_workspace != Workspaces::instance()->getActive() &&
+                        !isSticky()) {
                     Workspaces::instance()->setWorkspace(_workspace, false);
                 }
                 mapWindow();
@@ -2120,42 +2169,129 @@ Frame::handleClientMessage(XClientMessageEvent *ev, Client *client)
             raise();
             giveInputFocus();
         }
-    } else if (ev->message_type == ewmh->getAtom(NET_CLOSE_WINDOW)) {
+    } else if (ev->message_type == Atoms::getAtom(NET_CLOSE_WINDOW)) {
         client->close();
-    } else if (ev->message_type == ewmh->getAtom(NET_WM_DESKTOP)) {
+    } else if (ev->message_type == Atoms::getAtom(NET_WM_DESKTOP)) {
         if (client == _client) {
             setWorkspace(ev->data.l[0]);
         }
-    } else if (ev->message_type == IcccmAtoms::instance()->getAtom(WM_CHANGE_STATE) &&
+    } else if (ev->message_type == Atoms::getAtom(WM_CHANGE_STATE) &&
                (ev->format == 32) && (ev->data.l[0] == IconicState)) {
         if (client == _client) {
             iconify();
         }
     }
-#undef IS_STATE
 }
 
+/**
+ * Handle _NET_WM_STATE atom.
+ */
+void
+Frame::handleClientStateMessage(XClientMessageEvent *ev, Client *client)
+{
+    StateAction sa = getStateActionFromMessage(ev);
+    handleStateAtom(sa, ev->data.l[1], client);
+    if (ev->data.l[2] != 0) {
+        handleStateAtom(sa, ev->data.l[2], client);
+    }
+    client->updateEwmhStates();
+}
+
+/**
+ * Get StateAction from NET_WM atom.
+ */
+StateAction
+Frame::getStateActionFromMessage(XClientMessageEvent *ev)
+{
+    StateAction sa = STATE_SET;
+    if (ev->data.l[0]== NET_WM_STATE_REMOVE) {
+        sa = STATE_UNSET;
+    } else if (ev->data.l[0]== NET_WM_STATE_ADD) {
+        sa = STATE_SET;
+    } else if (ev->data.l[0]== NET_WM_STATE_TOGGLE) {
+        sa = STATE_TOGGLE;
+    }
+    return sa;
+}
+
+/**
+ * Handle state atom for client.
+ */
+void
+Frame::handleStateAtom(StateAction sa, Atom atom, Client *client)
+{
+    if (client == _client) {
+        handleCurrentClientStateAtom(sa, atom, client);
+    }
+
+    switch (atom) {
+    case STATE_SKIP_TASKBAR:
+        client->setStateSkip(sa, SKIP_TASKBAR);
+        break;
+    case STATE_SKIP_PAGER:
+        client->setStateSkip(sa, SKIP_PAGER);
+        break;
+    case STATE_DEMANDS_ATTENTION:
+        client->setStateDemandsAttention(sa, true);
+        break;
+    }
+}
+
+/**
+ * Handle state atom for actions that apply only on active client.
+ */
+void
+Frame::handleCurrentClientStateAtom(StateAction sa, Atom atom, Client *client)
+{
+    if (atom == Atoms::getAtom(STATE_STICKY)) {
+        setStateSticky(sa);
+    }
+    if (atom == Atoms::getAtom(STATE_MAXIMIZED_HORZ)
+            && ! client->isCfgDeny(CFG_DENY_STATE_MAXIMIZED_HORZ)) {
+        setStateMaximized(sa, true, false, false);
+    }
+    if (atom == Atoms::getAtom(STATE_MAXIMIZED_VERT)
+            && ! client->isCfgDeny(CFG_DENY_STATE_MAXIMIZED_VERT)) {
+        setStateMaximized(sa, false, true, false);
+    }
+    if (atom == Atoms::getAtom(STATE_SHADED)) {
+        setShaded(sa);
+    }
+    if (atom == Atoms::getAtom(STATE_HIDDEN)
+            && ! client->isCfgDeny(CFG_DENY_STATE_HIDDEN)) {
+        setStateIconified(sa);
+    }
+    if (atom == Atoms::getAtom(STATE_FULLSCREEN)
+            && ! client->isCfgDeny(CFG_DENY_STATE_FULLSCREEN)) {
+        setStateFullscreen(sa);
+    }
+    if (atom == Atoms::getAtom(STATE_ABOVE)
+            && ! client->isCfgDeny(CFG_DENY_STATE_ABOVE)) {
+        setStateAlwaysOnTop(sa);
+    }
+    if (atom == Atoms::getAtom(STATE_BELOW)
+            && ! client->isCfgDeny(CFG_DENY_STATE_BELOW)) {
+        setStateAlwaysBelow(sa);
+    }
+}
 
 //! @brief
 void
 Frame::handlePropertyChange(XPropertyEvent *ev, Client *client)
 {
-    EwmhAtoms *ewmh = EwmhAtoms::instance(); // convenience
-
-    if (ev->atom == ewmh->getAtom(NET_WM_DESKTOP)) {
+    if (ev->atom == Atoms::getAtom(NET_WM_DESKTOP)) {
         if (client == _client) {
             long workspace;
 
             if (AtomUtil::getLong(client->getWindow(),
-                                  ewmh->getAtom(NET_WM_DESKTOP), workspace)) {
+                                  Atoms::getAtom(NET_WM_DESKTOP), workspace)) {
                 if (workspace != signed(_workspace))
                     setWorkspace(workspace);
             }
         }
-    } else if (ev->atom == ewmh->getAtom(NET_WM_STRUT)) {
+    } else if (ev->atom == Atoms::getAtom(NET_WM_STRUT)) {
         client->getStrutHint();
-    } else if (ev->atom == ewmh->getAtom(NET_WM_NAME)
-               || ev->atom == XA_WM_NAME) {
+    } else if (ev->atom == Atoms::getAtom(NET_WM_NAME) || ev->atom == XA_WM_NAME) {
         handleTitleChange(client);
     } else if (ev->atom == XA_WM_NORMAL_HINTS) {
         client->getWMNormalHints();
@@ -2172,7 +2308,7 @@ void
 Frame::handleTitleChange(Client *client)
 {
     // Update title
-    client->getXClientName();
+    client->readName();
 
     bool require_render = true;
     if (client == _client) {
