@@ -1,13 +1,14 @@
 //
 // Harbour.cc for pekwm
-// Copyright (C) 2003-2004 Claes Nasten <pekdon{@}pekdon{.}net>
+// Copyright © 2003-2009 Claes Nästen <me@pekdon.net>
 //
 // This program is licensed under the GNU GPL.
 // See the LICENSE file for more information.
 //
 
-#include "../config.h"
-#ifdef HARBOUR
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif // HAVE_CONFIG_H
 
 #include "Harbour.hh"
 
@@ -17,11 +18,9 @@
 #include "DockApp.hh"
 #include "Workspaces.hh"
 #include "AutoProperties.hh"
-#ifdef MENUS
 #include "PDecor.hh"
 #include "PMenu.hh"
 #include "HarbourMenu.hh"
-#endif // MENUS
 
 #include <algorithm>
 #include <functional>
@@ -38,19 +37,14 @@ using std::endl;
 //! @brief Harbour constructor
 Harbour::Harbour(PScreen *s, Theme *t, Workspaces *w) :
         _scr(s), _theme(t), _workspaces(w),
-#ifdef MENUS
         _harbour_menu(0),
-#endif // MENUS
         _hidden(false), _size(0), _strut(0),
         _last_button_x(0), _last_button_y(0)
 {
     _strut = new Strut();
     _scr->addStrut(_strut);
     _strut->head = Config::instance()->getHarbourHead();
-
-#ifdef MENUS
     _harbour_menu = new HarbourMenu(_scr, _theme, this);
-#endif // MENUS
 }
 
 //! @brief Harbour destructor
@@ -60,12 +54,7 @@ Harbour::~Harbour(void)
 
     _scr->removeStrut(_strut);
     delete _strut;
-
-#ifdef MENUS
-    if (_harbour_menu) {
-        delete _harbour_menu;
-    }
-#endif // MENUS
+    delete _harbour_menu;
 }
 
 //! @brief Adds a DockApp to the Harbour
@@ -93,6 +82,9 @@ Harbour::addDockApp(DockApp *da)
         da->mapWindow();
     }
 
+#ifdef OPACITY
+    da->setOpacity(Config::instance()->getHarbourOpacity());
+#endif // OPACITY
     updateHarbourSize();
 }
 
@@ -181,7 +173,9 @@ void
 Harbour::restack(void)
 {
     PScreen::instance()->removeStrut(_strut);
-    if (Config::instance()->isHarbourOntop()) {
+    if (Config::instance()->isHarbourOntop() ||
+            ! Config::instance()->isHarbourMaximizeOver()) {
+
         PScreen::instance()->addStrut(_strut);
     }
     uint l = Config::instance()->isHarbourOntop() ? LAYER_DOCK : LAYER_DESKTOP;
@@ -310,7 +304,6 @@ Harbour::handleButtonEvent(XButtonEvent* ev, DockApp* da)
     _last_button_x = ev->x;
     _last_button_y = ev->y;
 
-#ifdef MENUS
     // FIXME: Make configurable
     if (ev->type == ButtonPress) {
         if (ev->button == BUTTON3) {
@@ -324,7 +317,6 @@ Harbour::handleButtonEvent(XButtonEvent* ev, DockApp* da)
             _harbour_menu->unmapWindow();
         }
     }
-#endif // MENUS
 }
 
 //! @brief Initiates moving of a DockApp based on info from a XMotionEvent.
@@ -434,7 +426,7 @@ Harbour::placeDockApp(DockApp *da)
                     continue; // exclude ourselves
                 }
                 
-                if (((*it)->getX() < signed(test + da->getWidth())) &&
+                if (((*it)->getX() < static_cast<signed>(test + da->getWidth())) &&
                         ((*it)->getRX() > test)) {
                     placed = increase = false;
                     test = right ? (*it)->getX() - da->getWidth() : (*it)->getRX();
@@ -463,7 +455,7 @@ Harbour::placeDockApp(DockApp *da)
                     continue; // exclude ourselves
                 }
                 
-                if (((*it)->getY() < signed(test + da->getHeight())) &&
+                if (((*it)->getY() < static_cast<signed>(test + da->getHeight())) &&
                         ((*it)->getBY() > test)) {
                     placed = increase = false;
                     test = right ? (*it)->getY() - da->getHeight() : (*it)->getBY();
@@ -522,7 +514,10 @@ Harbour::placeDockAppsSorted(void)
     }
 }
 
-//! @brief
+/**
+ * Make sure dock app is inside screen boundaries and placed on the
+ * right edge, usually called after resizing the dockapp.
+ */
 void
 Harbour::placeDockAppInsideScreen(DockApp *da)
 {
@@ -535,14 +530,14 @@ Harbour::placeDockAppInsideScreen(DockApp *da)
         // check horizontal position
         if (da->getX() < head.x) {
             da->move(head.x, da->getY());
-        } else if (da->getRX() > signed(head.x + head.width)) {
+        } else if (da->getRX() > static_cast<signed>(head.x + head.width)) {
             da->move(head.x + head.width - da->getWidth(), da->getY());
         }
 
         // check vertical position
-        if ((pos == TOP) && (da->getY() < head.y)) {
+        if ((pos == TOP) && (da->getY() != head.y)) {
             da->move(da->getX(), head.y);
-        } else if ((pos == BOTTOM) && (da->getBY() > signed(head.y + head.height))) {
+        } else if ((pos == BOTTOM) && (da->getBY() != static_cast<signed>(head.y + head.height))) {
             da->move(da->getX(), head.y + head.height - da->getHeight());
         }
 
@@ -551,14 +546,14 @@ Harbour::placeDockAppInsideScreen(DockApp *da)
         // check vertical position
         if (da->getY() < head.y) {
             da->move(da->getX(), head.y);
-        } else if (da->getBY() > signed(head.y + head.height)) {
+        } else if (da->getBY() > static_cast<signed>(head.y + head.height)) {
             da->move(da->getX(), head.y + head.height - da->getHeight());
         }
 
         // check horizontal position
-        if ((pos == LEFT) && (da->getX() < head.x)) {
+        if ((pos == LEFT) && (da->getX() != head.x)) {
             da->move(head.x, da->getY());
-        } else if ((pos == RIGHT) && (da->getRX() > signed(head.x + head.width))) {
+        } else if ((pos == RIGHT) && (da->getRX() != static_cast<signed>(head.x + head.width))) {
             da->move(head.x + head.width - da->getWidth(), da->getY());
         }
     }
@@ -633,4 +628,3 @@ Harbour::insertDockAppSorted(DockApp *da)
     _da_list.insert (it, da);
 }
 
-#endif // HARBOUR
