@@ -1,14 +1,14 @@
 //
 // HarbourMenu.cc for pekwm
-// Copyright (C) 2003-2004 Claes Nasten <pekdon{@}pekdon{.}net>
+// Copyright © 2003-2009 Claes Nästen <me@pekdon.net>
 //
 // This program is licensed under the GNU GPL.
 // See the LICENSE file for more information.
 //
 
-#include "../config.h"
-#ifdef HARBOUR
-#ifdef MENUS
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif // HAVE_CONFIG_H
 
 #include "PWinObj.hh"
 #include "PDecor.hh"
@@ -20,7 +20,7 @@
 
 //! @brief HarbourMenu constructor
 HarbourMenu::HarbourMenu(PScreen *scr, Theme *theme, Harbour *harbour)
-    : PMenu(scr->getDpy(), theme, L"Harbour", "HARBOUR"),
+    : PMenu(theme, L"Harbour", "HARBOUR"),
       _harbour(harbour), _dockapp(0)
 {
     ActionEvent ae;
@@ -60,5 +60,3 @@ HarbourMenu::handleItemExec(PMenu::Item *item)
     }
 }
 
-#endif // MENUS
-#endif // HARBOUR
