@@ -1,6 +1,6 @@
 //
 // CmdDialog.hh for pekwm
-// Copyright © 2008 Claes Nästén <me@pekdon.net>
+// Copyright © 2009 Claes Nästén <me@pekdon.net>
 //
 // This program is licensed under the GNU GPL.
 // See the LICENSE file for more information.
@@ -11,6 +11,8 @@
 #endif // HAVE_CONFIG_H
 
 #include <list>
+#include <cwchar>
+#include <cwctype>
 
 #include "InputDialog.hh"
 #include "KeyGrabber.hh"
@@ -24,20 +26,28 @@ extern "C" {
 }
 
 using std::list;
+using std::map;
 using std::wstring;
+using std::iswprint;
+
+map<KeySym, wchar_t> InputDialog::_keysym_map;
 
 /**
  * InputDialog constructor.
  */
-InputDialog::InputDialog(Display *dpy, Theme *theme, const std::wstring &title)
-  : PDecor(dpy, theme, "INPUTDIALOG"),
-    _data(theme->getCmdDialogData()),
-    _pixmap_bg(None), _pos(0), _buf_off(0), _buf_chars(0)
+InputDialog::InputDialog(Theme *theme, const std::wstring &title)
+    : PDecor(theme, "INPUTDIALOG"), PWinObjReference(0),
+      _data(theme->getCmdDialogData()),
+      _pixmap_bg(None), _pos(0), _buf_off(0), _buf_chars(0)
 {
     // PWinObj attributes
-    _layer = LAYER_NONE; // hack, goes over LAYER_MENU
+    setLayer(LAYER_NONE); // hack, goes over LAYER_MENU
     _hidden = true; // don't care about it when changing worskpace etc
 
+    if (! _keysym_map.size()) {
+        reloadKeysymMap();
+    }
+
     // Add action to list, going to be used from close and exec
     ::Action action;
     _ae.action_list.push_back(action);
@@ -46,7 +56,7 @@ InputDialog::InputDialog(Display *dpy, Theme *theme, const std::wstring &title)
     titleSetActive(0);
     setTitle(title);
 
-    _text_wo = new PWinObj(_dpy);
+    _text_wo = new PWinObj;
     XSetWindowAttributes attr;
     attr.override_redirect = false;
     attr.event_mask = ButtonPressMask|ButtonReleaseMask|ButtonMotionMask|
@@ -89,6 +99,41 @@ InputDialog::~InputDialog(void)
     unloadTheme();
 }
 
+void
+InputDialog::reloadKeysymMap(void)
+{
+    _keysym_map.clear();
+
+    addKeysymToKeysymMap(XK_KP_0, L'0');
+    addKeysymToKeysymMap(XK_KP_1, L'1');
+    addKeysymToKeysymMap(XK_KP_2, L'2');
+    addKeysymToKeysymMap(XK_KP_3, L'3');
+    addKeysymToKeysymMap(XK_KP_4, L'4');
+    addKeysymToKeysymMap(XK_KP_5, L'5');
+    addKeysymToKeysymMap(XK_KP_6, L'6');
+    addKeysymToKeysymMap(XK_KP_7, L'7');
+    addKeysymToKeysymMap(XK_KP_8, L'8');
+    addKeysymToKeysymMap(XK_KP_9, L'9');
+}
+
+void
+InputDialog::addKeysymToKeysymMap(KeySym keysym, wchar_t chr)
+{
+    Display *dpy = PScreen::instance()->getDpy();
+
+    int keysyms_per_keycode;
+    KeyCode keycode = XKeysymToKeycode(dpy, keysym);
+    KeySym *keysyms = XGetKeyboardMapping(dpy, keycode, 1, &keysyms_per_keycode);
+
+    for (int i = 0; i < keysyms_per_keycode; i++) {
+        if (keysyms[i] != NoSymbol) {
+            _keysym_map[keysyms[i]] = chr;
+        }
+    }
+
+    XFree(keysyms);
+}
+
 /**
  * Handles ButtonPress, moving the text cursor
  */
@@ -117,15 +162,18 @@ InputDialog::handleKeyPress(XKeyEvent *ev)
             switch (it->getAction()) {
             case INPUT_INSERT:
                 bufAdd(ev);
+                completeReset();
                 break;
             case INPUT_REMOVE:
                 bufRemove();
                 break;
             case INPUT_CLEAR:
                 bufClear();
+                completeReset();
                 break;
             case INPUT_CLEARFROMCURSOR:
                 bufKill();
+                completeReset();
                 break;
             case INPUT_EXEC:
                 ae = exec();
@@ -134,24 +182,34 @@ InputDialog::handleKeyPress(XKeyEvent *ev)
                 ae = close();
                 break;
             case INPUT_COMPLETE:
+                complete();
+                break;
+            case INPUT_COMPLETE_ABORT:
+                completeAbort();
                 break;
             case INPUT_CURS_NEXT:
                 bufChangePos(1);
+                completeReset();
                 break;
             case INPUT_CURS_PREV:
                 bufChangePos(-1);
+                completeReset();
                 break;
             case INPUT_CURS_BEGIN:
                 _pos = 0;
+                completeReset();
                 break;
             case INPUT_CURS_END:
                 _pos = _buf.size();
+                completeReset();
                 break;
             case INPUT_HIST_NEXT:
                 histNext();
+                completeReset();
                 break;
             case INPUT_HIST_PREV:
                 histPrev();
+                completeReset();
                 break;
             case INPUT_NO_ACTION:
             default:
@@ -331,22 +389,62 @@ InputDialog::close(void)
     return &_ae;
 }
 
+/**
+ * Default implementation, fills in _buf_on_complete to safely run
+ * completeAbort.
+ */
+void
+InputDialog::complete(void)
+{
+    _buf_on_complete = _buf;
+    _pos_on_complete = _pos;
+}
+
+/**
+ * Restore buffer and clear completion buffers.
+ */
+void
+InputDialog::completeAbort(void)
+{
+    if (_buf_on_complete.size()) {
+        _buf = _buf_on_complete;
+        _pos = _pos_on_complete;
+    }
+
+    completeReset();
+}
+
+/**
+ * Clear the completion buffer.
+ */
+void
+InputDialog::completeReset(void)
+{
+    // Old gcc doesn't know about .clear()
+    _buf_on_complete = _buf_on_complete_result = L"";
+    _pos_on_complete = 0;
+}
+
 /**
  * Adds char to buffer
  */
 void
 InputDialog::bufAdd(XKeyEvent *ev)
 {
+    KeySym keysym;
     char c_return[64];
-    memset(c_return, '\0', 64);
+    memset(c_return, '\0', sizeof(c_return));
 
-    XLookupString(ev, c_return, 64, 0, 0);
-
-    // Add wide string to buffer counting position
-    wstring buf_ret(Util::to_wide_str(c_return));
-    for (unsigned int i = 0; i < buf_ret.size(); ++i) {
-        if (iswprint(buf_ret[i])) {
-            _buf.insert(_buf.begin() + _pos++, buf_ret[i]);
+    XLookupString(ev, c_return, sizeof(c_return), &keysym, 0);
+    if (_keysym_map.count(keysym)) {
+        _buf.insert(_buf.begin() + _pos++, _keysym_map[keysym]);
+    } else {
+        // Add wide string to buffer counting position
+        wstring buf_ret(Util::to_wide_str(c_return));
+        for (unsigned int i = 0; i < buf_ret.size(); ++i) {
+            if (iswprint(buf_ret[i])) {
+                _buf.insert(_buf.begin() + _pos++, buf_ret[i]);
+            }
         }
     }
 }
@@ -370,8 +468,8 @@ InputDialog::bufRemove(void)
 void
 InputDialog::bufClear(void)
 {
-    _buf = L""; // old gcc doesn't know about .clear()
-    _pos = _buf_off = _buf_chars = 0;
+    _buf = _buf_on_complete = _buf_on_complete_result = L""; // old gcc doesn't know about .clear()
+    _pos = _pos_on_complete = _buf_off = _buf_chars = 0;
 }
 
 /**
@@ -477,12 +575,12 @@ InputDialog::histPrev(void)
 void
 InputDialog::updateSize(void)
 {
-    Geometry head;
-    PScreen::instance()->getHeadInfo(PScreen::instance()->getNearestHead(_gm.x, _gm.y), head);
-
     // Resize the child window and update the size depending.
     uint old_width = _gm.width;
-    resizeChild(head.width / 3, _data->getFont()->getHeight() + _data->getPad(PAD_UP) + _data->getPad(PAD_DOWN));
+
+    unsigned int width, height;
+    getInputSize(width, height);
+    resizeChild(width, height);
 
     // If size was updated, replace the texture and recalculate display
     // buffer.
@@ -507,3 +605,19 @@ InputDialog::updatePixmapSize(void)
     _text_wo->setBackgroundPixmap(_pixmap_bg);
     _text_wo->clear();
 }
+
+/**
+ * Get size of the text input widget.
+ *
+ * @param width Fill in width.
+ * @param height Fill in height.
+ */
+void
+InputDialog::getInputSize(unsigned int &width, unsigned int &height)
+{
+    Geometry head;
+    PScreen::instance()->getHeadInfo(PScreen::instance()->getNearestHead(_gm.x, _gm.y), head);
+
+    width = head.width / 3;
+    height = _data->getFont()->getHeight() + _data->getPad(PAD_UP) + _data->getPad(PAD_DOWN);
+}
