@@ -1,6 +1,6 @@
 //
 // PMenu.cc for pekwm
-// Copyright © 2004-2008 Claes Nästén <me@pekdon.net>
+// Copyright © 2004-2009 Claes Nästén <me@pekdon.net>
 //
 // This program is licensed under the GNU GPL.
 // See the LICENSE file for more information.
@@ -24,6 +24,7 @@
 #include "Theme.hh"
 #include "PixmapHandler.hh"
 #include "Workspaces.hh"
+#include "AutoProperties.hh"
 
 #include <algorithm>
 #include <cstdlib>
@@ -37,11 +38,13 @@ using std::endl;
 using std::list;
 using std::map;
 using std::string;
+using std::wstring;
 using std::find;
 
 PMenu::Item::Item(const std::wstring &name, PWinObj *wo_ref, PTexture *icon)
-  : _x(0), _y(0), _name(name), _wo_ref(wo_ref),
-    _type(MENU_ITEM_NORMAL), _icon(icon), _dynamic(false)
+    : PWinObjReference(wo_ref),
+      _x(0), _y(0), _name(name), 
+      _type(MENU_ITEM_NORMAL), _icon(icon), _creator(0)
 {
     if (_icon) {
         TextureHandler::instance()->referenceTexture(_icon);
@@ -58,12 +61,14 @@ PMenu::Item::~Item(void)
 map<Window,PMenu*> PMenu::_menu_map = map<Window,PMenu*>();
 
 //! @brief Constructor for PMenu class
-PMenu::PMenu(Display *dpy, Theme *theme, const std::wstring &title,
+PMenu::PMenu(Theme *theme, const std::wstring &title,
              const std::string &name, const std::string decor_name)
-    : PDecor(dpy, theme, decor_name),
+    : PDecor(theme, decor_name),
       _name(name),
-      _menu_parent(0), _menu_wo(0),
+      _menu_parent(0), _class_hint(L"pekwm", L"Menu", L"", L"", L""),
+      _menu_wo(0),
       _menu_bg_fo(None), _menu_bg_un(None), _menu_bg_se(None),
+      _menu_width(0),
       _item_height(0), _item_width_max(0), _item_width_max_avail(0),
       _icon_width(0), _icon_height(0),
       _separator_height(0),
@@ -74,15 +79,11 @@ PMenu::PMenu(Display *dpy, Theme *theme, const std::wstring &title,
 
     // PWinObj attributes
     _type = PWinObj::WO_MENU;
-    _layer = LAYER_MENU;
+    setLayer(LAYER_MENU);
     _hidden = true; // don't care about it when changing worskpace etc
 
-    titleAdd(&_title);
-    titleSetActive(0);
-    setTitle(title);
-
     // create menu content child
-    _menu_wo = new PWinObj(_dpy);
+    _menu_wo = new PWinObj;
     XSetWindowAttributes attr;
     attr.override_redirect = True;
     attr.event_mask = ButtonPressMask|ButtonReleaseMask|ButtonMotionMask|
@@ -91,6 +92,11 @@ PMenu::PMenu(Display *dpy, Theme *theme, const std::wstring &title,
                                       0, 0, 1, 1, 0,
                                       CopyFromParent, InputOutput, CopyFromParent,
                                       CWOverrideRedirect|CWEventMask, &attr));
+
+    titleAdd(&_title);
+    titleSetActive(0);
+    setTitle(title);
+	
     addChild(_menu_wo);
     addChildWindow(_menu_wo->getWindow());
     activateChild(_menu_wo);
@@ -100,6 +106,10 @@ PMenu::PMenu(Display *dpy, Theme *theme, const std::wstring &title,
     _menu_map[_window] = this; // add to menu map
     woListAdd(this);
     _wo_map[_window] = this;
+#ifdef OPACITY
+    setOpacity(Config::instance()->getMenuFocusOpacity(),
+               Config::instance()->getMenuUnfocusOpacity());
+#endif // OPACITY
 }
 
 //! @brief Destructor for PMenu class
@@ -157,6 +167,16 @@ PMenu::setFocused(bool focused)
     }
 }
 
+/**
+ * Set focusable, includes the _menu_wo as well as the decor.
+ */
+void
+PMenu::setFocusable(bool focusable)
+{
+    PDecor::setFocusable(focusable);
+    _menu_wo->setFocusable(focusable);
+}
+
 //! @brief
 ActionEvent*
 PMenu::handleButtonPress(XButtonEvent *ev)
@@ -176,10 +196,18 @@ PMenu::handleButtonPress(XButtonEvent *ev)
     }
 }
 
-//! @brief
+/**
+ * Handle button release.
+ */
 ActionEvent*
 PMenu::handleButtonRelease(XButtonEvent *ev)
 {
+    if (_window == ev->subwindow) {
+        ev->window = _menu_wo->getWindow();
+        ev->x -= _gm.x;
+        ev->y -= _gm.y + getTitleHeight();
+    }
+
     if (*_menu_wo == ev->window) {
         MouseEventType mb = MOUSE_EVENT_RELEASE;
 
@@ -205,17 +233,27 @@ PMenu::handleButtonRelease(XButtonEvent *ev)
     }
 }
 
-//! @brief
+/**
+ * Handle motion event, select buttons and execute actions.
+ *
+ * @param ev Handle motion event.
+ */
 ActionEvent*
 PMenu::handleMotionEvent(XMotionEvent *ev)
 {
+    if (_window == ev->subwindow) {
+        ev->window = _menu_wo->getWindow();
+        ev->x -= _gm.x;
+        ev->y -= _gm.y + getTitleHeight();
+    }
+
     if (*_menu_wo == ev->window) {
-        handleItemEvent(MOUSE_EVENT_MOTION, ev->x, ev->y);
+        uint button = PScreen::instance()->getButtonFromState(ev->state);
+        handleItemEvent(button ? MOUSE_EVENT_MOTION_PRESSED : MOUSE_EVENT_MOTION, ev->x, ev->y);
 
         ActionEvent *ae;
-        uint button = PScreen::instance()->getButtonFromState(ev->state);
-        ae = ActionHandler::findMouseAction(button, ev->state,
-                                            MOUSE_EVENT_MOTION,
+        PScreen::stripButtonModifiers(&ev->state);
+        ae = ActionHandler::findMouseAction(button, ev->state, MOUSE_EVENT_MOTION,
                                             Config::instance()->getMouseActionList(MOUSE_ACTION_LIST_MENU));
 
         // check motion threshold
@@ -236,8 +274,7 @@ ActionEvent*
 PMenu::handleEnterEvent(XCrossingEvent *ev)
 {
     if (*_menu_wo == ev->window) {
-        return ActionHandler::findMouseAction(BUTTON_ANY, ev->state,
-                                              MOUSE_EVENT_ENTER,
+        return ActionHandler::findMouseAction(BUTTON_ANY, ev->state, MOUSE_EVENT_ENTER,
                                               Config::instance()->getMouseActionList(MOUSE_ACTION_LIST_MENU));
     } else {
         return PDecor::handleEnterEvent(ev);
@@ -272,25 +309,30 @@ PMenu::loadTheme(void)
 
 // END - PDecor interface.
 
-//! @brief
+/**
+ * Handle event on menu item at position, ignores event if no item
+ * exists at the position.
+ */
 void
 PMenu::handleItemEvent(MouseEventType type, int x, int y)
 {
     PMenu::Item *item = findItem(x, y);
-    if (! item)
+    if (! item) {
         return;
+    }
 
     // Unmap submenu if we enter them on the same event as selecting.
     if (((_item_curr == _item_list.end()) || (item != *_item_curr))
-            && Config::instance()->isMenuSelectOn(type))
+        && Config::instance()->isMenuSelectOn(type)) {
         select(item, Config::instance()->isMenuEnterOn(type));
+    }
 
     if (Config::instance()->isMenuEnterOn(type)) {
-        if (item->getWORef() &&
-                (item->getWORef()->getType() == PWinObj::WO_MENU)) {
-            // special case for motion, would flicker like crazy if we didn't check
-            if ((type != MOUSE_EVENT_MOTION) &&
-                    item->getWORef()->isMapped()) {
+        if (item->getWORef()
+            && (item->getWORef()->getType() == PWinObj::WO_MENU)) {
+            // Special case for motion, would flicker like crazy if we didn't check
+            if ((type != MOUSE_EVENT_MOTION) && (type != MOUSE_EVENT_MOTION_PRESSED)
+                && item->getWORef()->isMapped()) {
                 static_cast<PMenu*>(item->getWORef())->unmapSubmenus();
                 item->getWORef()->unmapWindow();
 
@@ -303,8 +345,7 @@ PMenu::handleItemEvent(MouseEventType type, int x, int y)
     }
 
     // Submenus don't have any actions, so we don't exec ( and close them )
-    if (item->getAE().action_list.size()
-            && Config::instance()->isMenuExecOn(type)) {
+    if (item->getAE().action_list.size() && Config::instance()->isMenuExecOn(type)) {
         exec(item);
     }
 }
@@ -330,13 +371,12 @@ PMenu::buildMenu(void)
 void
 PMenu::buildMenuCalculate(void)
 {
-    uint width = 1, height = 1, sep = 0;
-    list<PMenu::Item*>::iterator it;
     _has_submenu = false;
 
     // Get how many visible objects we have
-    _size = 0;
-    for (it = _item_list.begin(); it != _item_list.end(); ++it) {
+    unsigned int sep = 0;
+    list<PMenu::Item*>::iterator it(_item_list.begin());
+    for (_size = 0; it != _item_list.end(); ++it) {
         if ((*it)->getType() == PMenu::Item::MENU_ITEM_NORMAL) {
             ++_size;
         } else if ((*it)->getType() == PMenu::Item::MENU_ITEM_SEPARATOR) {
@@ -345,18 +385,74 @@ PMenu::buildMenuCalculate(void)
     }
 
     if (_size == 0) {
-#ifdef DEBUG
-        cerr << __FILE__ << "@" << __LINE__ << ": "
-             << "PMenu(" << this << ")::buildMenuCalculate()" << endl
-             << " *** _size == 0" << endl;
-#endif // DEBUG
         return;
     }
 
+    unsigned int width = 1, height = 1;
+    buildMenuCalculateMaxWidth(width, height);
+
+    // FIXME: Remove extra padding from calculation
+    if (_menu_width) {
+      _item_width_max = _menu_width;
+    }
+
+    // This is the available width for drawing text on, the rest is reserved
+    // for submenu indicator, padding etc.
+    _item_width_max_avail = _item_width_max;
+
+    // Continue add padding etc.
+    _item_width_max += _theme->getMenuData()->getPad(PAD_LEFT)
+        + _theme->getMenuData()->getPad(PAD_RIGHT);
+    if (Config::instance()->isDisplayMenuIcons()) {
+        _item_width_max += _icon_width;
+    }
+
+    // If we have any submenus, increase the maximum width with arrow width +
+    // right pad as we are going to pad the arrow from the text too.
+    if (_has_submenu) {
+        _item_width_max += _theme->getMenuData()->getPad(PAD_RIGHT)
+            + _theme->getMenuData()->getTextureArrow(OBJECT_STATE_FOCUSED)->getWidth();
+    }
+
+    // Remove padding etc from avail and item width.
+    if (_menu_width) {
+        unsigned int padding = _item_width_max - _item_width_max_avail;
+        _item_width_max -= padding;
+        _item_width_max_avail -= padding;
+    }
+
+    // Calculate item height
+    _item_height = std::max(_theme->getMenuData()->getFont(OBJECT_STATE_FOCUSED)->getHeight(), _icon_height)
+      + _theme->getMenuData()->getPad(PAD_UP)
+      + _theme->getMenuData()->getPad(PAD_DOWN);
+    _separator_height = _theme->getMenuData()->getTextureSeparator(OBJECT_STATE_FOCUSED)->getHeight();
+
+    height = (_item_height * _size) + (_separator_height * sep);
+
+    if (_size) {
+    	_size += sep;
+    }
+
+    buildMenuCalculateColumns(width, height);
+
+    // Check if we need to enable scrolling
+    _scroll = (width > PScreen::instance()->getWidth());
+
+    resizeChild(width, height);
+}
+
+/**
+ * Get maximum item width and icon size.
+ */
+void
+PMenu::buildMenuCalculateMaxWidth(unsigned int &width, unsigned int &height)
+{
     // Calculate max item width, to be used if/when splitting a menu
     // up in rows because of limited vertical space.
     _item_width_max = 1;
     _icon_width = _icon_height = 0;
+
+    list<PMenu::Item*>::iterator it(_item_list.begin());
     for (it = _item_list.begin(); it != _item_list.end(); ++it) {
         // Only include standard items
         if ((*it)->getType() != PMenu::Item::MENU_ITEM_NORMAL) {
@@ -385,50 +481,41 @@ PMenu::buildMenuCalculate(void)
         }
     }
 
+
     // Make sure icon width and height are not larger than configured.
-    if (Config::instance()->getMenuIconWidth()) {
-      _icon_width = std::min(Config::instance()->getMenuIconWidth(), _icon_width);
-    }
-    if (Config::instance()->getMenuIconHeight()) {
-      _icon_height = std::min(Config::instance()->getMenuIconHeight(), _icon_height);
+    if (_icon_width) {
+        _icon_width = Util::between<uint>(_icon_width,
+                                          Config::instance()->getMenuIconLimit(_icon_width, WIDTH_MIN, _name),
+                                          Config::instance()->getMenuIconLimit(_icon_width, WIDTH_MAX, _name));
+        _icon_height = Util::between<uint>(_icon_height,
+                                           Config::instance()->getMenuIconLimit(_icon_height, HEIGHT_MIN, _name),
+                                           Config::instance()->getMenuIconLimit(_icon_height, HEIGHT_MAX, _name));
     }
+}
 
-    // This is the available width for drawing text on, the rest is reserved
-    // for submenu indicator, padding etc.
-    _item_width_max_avail = _item_width_max;
-
-    // Continue add padding etc.
-    _item_width_max += _theme->getMenuData()->getPad(PAD_LEFT)
-      + _theme->getMenuData()->getPad(PAD_RIGHT) + _icon_width;
-
-    // If we have any submenus, increase the maximum width with arrow width +
-    // right pad as we are going to pad the arrow from the text too.
-    if (_has_submenu) {
-        _item_width_max += _theme->getMenuData()->getPad(PAD_RIGHT)
-                           + _theme->getMenuData()->getTextureArrow(OBJECT_STATE_FOCUSED)->getWidth();
-    }
-
-    // calculate item height
-    _item_height = std::max(_theme->getMenuData()->getFont(OBJECT_STATE_FOCUSED)->getHeight(), _icon_height)
-      + _theme->getMenuData()->getPad(PAD_UP)
-      + _theme->getMenuData()->getPad(PAD_DOWN);
-    _separator_height = _theme->getMenuData()->getTextureSeparator(OBJECT_STATE_FOCUSED)->getHeight();
-
-    height = (_item_height * _size) + (_separator_height * sep);
-
-    // check if menu should have more than one row
-    if ((height + getTitleHeight()) > PScreen::instance()->getHeight()) {
-        _cols = height / (PScreen::instance()->getHeight() - getTitleHeight());
-        if ((height % (PScreen::instance()->getHeight() - getTitleHeight())) != 0) {
-            ++_cols;
-        }
-        _rows = _size / _cols;
-        if ((_size % _cols) != 0) {
-            ++_rows;
-        }
-    } else {
+/**
+ * Calculate number of columns, this does not apply to static width
+ * menus.
+ */
+void
+PMenu::buildMenuCalculateColumns(unsigned int &width, unsigned int &height)
+{
+    // Check if the menu fits or is static width
+    if (_menu_width
+        || (height + getTitleHeight()) <= PScreen::instance()->getHeight()) {
         _cols = 1;
+        width = _menu_width ? _menu_width : _item_width_max;
         _rows = _size;
+        return;
+    }
+
+    _cols = height / (PScreen::instance()->getHeight() - getTitleHeight());
+    if ((height % (PScreen::instance()->getHeight() - getTitleHeight())) != 0) {
+        ++_cols;
+    }
+    _rows = _size / _cols;
+    if ((_size % _cols) != 0) {
+        ++_rows;
     }
 
     width = _cols * _item_width_max;
@@ -436,15 +523,21 @@ PMenu::buildMenuCalculate(void)
     if (_cols > 1) {
         uint i, j, row_height;
         height = 0;
-
+       
+        list<PMenu::Item*>::iterator it(_item_list.begin());
         for (i = 0, it = _item_list.begin(); i < _cols; ++i) {
             row_height = 0;
-            for (j = 0; (j < _rows) && (it != _item_list.end()); ++it) {
-                if ((*it)->getType() == PMenu::Item::MENU_ITEM_NORMAL) {
+            for (j = 0; (j < _rows) && (it != _item_list.end()); ++it, ++j) {
+                switch ((*it)->getType()) {
+                case PMenu::Item::MENU_ITEM_NORMAL:
                     row_height += _item_height;
-                    ++j; // only count real menu items
-                } else if ((*it)->getType() == PMenu::Item::MENU_ITEM_SEPARATOR) {
+                    break;
+                case PMenu::Item::MENU_ITEM_SEPARATOR:
                     row_height += _separator_height;
+                    break;
+                case PMenu::Item::MENU_ITEM_HIDDEN:
+                default:
+                	break;
                 }
             }
 
@@ -453,15 +546,6 @@ PMenu::buildMenuCalculate(void)
             }
         }
     }
-
-    // check if we need to enable scrolling
-    if (width > PScreen::instance()->getWidth()) {
-        _scroll = true;
-    } else {
-        _scroll = false;
-    }
-
-    resizeChild(width, height);
 }
 
 //! @brief Places the items in the menu
@@ -542,50 +626,56 @@ PMenu::buildMenuRenderItem(Pixmap pix, ObjectState state, PMenu::Item *item)
 
     if (item->getType() == PMenu::Item::MENU_ITEM_NORMAL) {
         tex = md->getTextureItem(state);
-        tex->render(pix,
-                    item->getX(), item->getY(), _item_width_max, _item_height);
+        tex->render(pix, item->getX(), item->getY(), _item_width_max, _item_height);
 
+        uint start_x, start_y, icon_width, icon_height;
         // If entry has an icon, draw it
-        if (item->getIcon()) {
-          uint i_w = std::min(Config::instance()->getMenuIconWidth(),
-                              item->getIcon()->getWidth());
-          uint i_h = std::min(Config::instance()->getMenuIconHeight(),
-                              item->getIcon()->getHeight());
+        if (item->getIcon() && Config::instance()->isDisplayMenuIcons()) {
+            icon_width = Util::between<uint>(item->getIcon()->getWidth(),
+                                             Config::instance()->getMenuIconLimit(_icon_width, WIDTH_MIN, _name),
+                                             Config::instance()->getMenuIconLimit(_icon_width, WIDTH_MAX, _name));
+            
+            icon_height = Util::between<uint>(item->getIcon()->getHeight(),
+                                              Config::instance()->getMenuIconLimit(_icon_height, HEIGHT_MIN, _name),
+                                              Config::instance()->getMenuIconLimit(_icon_height, HEIGHT_MAX, _name));
 
-          item->getIcon()->render(pix,
-                                  item->getX() + md->getPad(PAD_LEFT) + (_icon_width - i_w) / 2,
-                                  item->getY() + (_item_height - i_h) / 2,
-                                  i_w, i_h);
+            start_x = item->getX() + md->getPad(PAD_LEFT) + (_icon_width - icon_width) / 2;
+            start_y = item->getY() + (_item_height - icon_height) / 2;
+            item->getIcon()->render(pix, start_x, start_y, icon_width, icon_height);
+        } else {
+            icon_width = 0;
+            icon_height = 0;
         }
 
         // If entry has a submenu, lets draw our submenu "arrow"
-        if (item->getWORef()
-                && (item->getWORef()->getType() == PWinObj::WO_MENU)) {
+        if (item->getWORef() && (item->getWORef()->getType() == PWinObj::WO_MENU)) {
             tex = md->getTextureArrow(state);
-            uint a_w = tex->getWidth();
-            uint a_h = tex->getHeight();
-            uint a_y = static_cast<uint>((_item_height / 2) - (a_h / 2));
+            uint arrow_width = tex->getWidth();
+            uint arrow_height = tex->getHeight();
+            uint arrow_y = static_cast<uint>((_item_height / 2) - (arrow_height / 2));
 
-            tex->render(pix,
-                        item->getX() + _item_width_max - a_w -
-                        md->getPad(PAD_RIGHT),
-                        item->getY() + a_y,
-                        a_w, a_h);
+            start_x = item->getX() + _item_width_max - arrow_width - md->getPad(PAD_RIGHT);
+            start_y = item->getY() + arrow_y;
+            tex->render(pix, start_x, start_y, arrow_width, arrow_height);
         }
 
         PFont *font = md->getFont(state);
-        font->draw(pix,
-                   item->getX() + md->getPad(PAD_LEFT) + _icon_width,
-                   item->getY() + md->getPad(PAD_UP),
-                   item->getName().c_str(), 0, // Text and max chars
-                   _item_width_max_avail);
+        start_x = item->getX() + md->getPad(PAD_LEFT);
+        // Add icon width to starting x position if frame icons are enabled.
+        if (Config::instance()->isDisplayMenuIcons()) {
+            start_x += _icon_width;
+        }
+		
+        start_y = item->getY() + md->getPad(PAD_UP)
+            + (_item_height - font->getHeight() - md->getPad(PAD_UP) - md->getPad(PAD_DOWN)) / 2;
+
+        // Render item text.
+        font->draw(pix, start_x, start_y, item->getName().c_str(), 0, _item_width_max_avail);
 
     } else if ((item->getType() == PMenu::Item::MENU_ITEM_SEPARATOR) &&
                (state < OBJECT_STATE_SELECTED)) {
         tex = md->getTextureSeparator(state);
-        tex->render(pix,
-                    item->getX(), item->getY(),
-                    _item_width_max, _separator_height);
+        tex->render(pix, item->getX(), item->getY(), _item_width_max, _separator_height);
     }
 }
 
@@ -600,15 +690,16 @@ PMenu::buildMenuRenderItem(Pixmap pix, ObjectState state, PMenu::Item *item)
 void
 PMenu::selectItem(std::list<PMenu::Item*>::iterator item, bool unmap_submenu)
 {
-    if (_item_curr == item)
+    if (_item_curr == item) {
         return;
+    }
 
     deselectItem(unmap_submenu);
-
     _item_curr = item;
 
-    if (_mapped)
+    if (_mapped) {
         COPY_ITEM_AREA((*item), _menu_bg_se);
+    }
 }
 
 //! @brief Deselects selected item
@@ -696,6 +787,26 @@ void
 PMenu::setTitle(const std::wstring &title)
 {
     _title.setReal(title);
+	
+    // Apply title rules to allow title rewriting
+    applyTitleRules(title);
+}
+
+/**
+ * Applies title rules to menu.
+ */
+void
+PMenu::applyTitleRules(const std::wstring &title)
+{
+    _class_hint.title = title;
+    TitleProperty *data = AutoProperties::instance()->findTitleProperty(&_class_hint);
+
+    if (data) {
+        wstring new_title(title);
+        if (data->getTitleRule().ed_s(new_title)) {
+            _title.setCustom(new_title);
+        }
+    }
 }
 
 //! @brief Inserts item into the menu ( without rebuilding )
@@ -747,8 +858,9 @@ PMenu::remove(PMenu::Item *item)
         return;
     }
 
-    if ((_item_curr != _item_list.end()) && (item == *_item_curr))
+    if ((_item_curr != _item_list.end()) && (item == *_item_curr)) {
         _item_curr = _item_list.end();
+    }
 
     delete item;
     _item_list.remove(item);
@@ -763,6 +875,7 @@ PMenu::removeAll(void)
         delete *it;
     }
     _item_list.clear();
+    _item_curr = _item_list.end();
 }
 
 //! @brief Places the menu under the mouse and maps it.
@@ -816,11 +929,9 @@ PMenu::unmapSubmenus(void)
     list<PMenu::Item*>::iterator it(_item_list.begin());
     for (; it != _item_list.end(); ++it) {
         if ((*it)->getWORef() && (*it)->getWORef()->getType() == PWinObj::WO_MENU) {
-            if (! (*it)->isDynamic()) {
-                // Sub-menus will be deleted when unmapping this, so no need
-                // to continue.
-                static_cast<PMenu*>((*it)->getWORef())->unmapSubmenus();
-            }
+            // Sub-menus will be deleted when unmapping this, so no need
+            // to continue.
+            static_cast<PMenu*>((*it)->getWORef())->unmapSubmenus();
             (*it)->getWORef()->unmapWindow();
         }
     }
