@@ -1,6 +1,6 @@
 //
 // PScreen.cc for pekwm
-// Copyright © 2003-2008 Claes Nästén <me{@}pekdon{.}net>
+// Copyright © 2003-2009 Claes Nästén <me@pekdon.net>
 //
 // This program is licensed under the GNU GPL.
 // See the LICENSE file for more information.
@@ -13,6 +13,7 @@
 #include <string>
 #include <iostream>
 #include <cassert>
+#include <cstring> // required for memset in FD_ZERO
 
 #ifdef HAVE_LIMITS
 #include <limits>
@@ -30,9 +31,18 @@ extern "C" {
 #endif // HAVE_XRANDR
 #include <X11/keysym.h> // For XK_ entries
 #include <sys/select.h>
+
+#ifdef DEBUG
+bool xerrors_ignore = false;
+#endif // DEBUG
+
+unsigned int xerrors_count = 0;
 }
 
 #include "PScreen.hh"
+// FIXME: Remove when strut handling is moved away from here.
+#include "PWinObj.hh"
+#include "ManagerWindows.hh"
 
 using std::cerr;
 using std::endl;
@@ -40,15 +50,40 @@ using std::vector;
 using std::list;
 using std::map;
 using std::string;
+using std::memset; // required for FD_ZERO
 
 const uint PScreen::MODIFIER_TO_MASK[] = {
     ShiftMask, LockMask, ControlMask,
     Mod1Mask, Mod2Mask, Mod3Mask, Mod4Mask, Mod5Mask
 };
-const uint PScreen::MODIFIER_TO_MASK_NUM = sizeof(PScreen::MODIFIER_TO_MASK[0]) / sizeof(PScreen::MODIFIER_TO_MASK);
+const uint PScreen::MODIFIER_TO_MASK_NUM = sizeof(PScreen::MODIFIER_TO_MASK) / sizeof(PScreen::MODIFIER_TO_MASK[0]);
 
 PScreen* PScreen::_instance = 0;
 
+
+extern "C" {
+    /**
+      * XError handler, prints error.
+      */
+    static int
+    handleXError(Display *dpy, XErrorEvent *ev)
+    {
+        ++xerrors_count;
+
+#ifdef DEBUG
+        if (xerrors_ignore) {
+            return 0;
+        }
+
+        char error_buf[256];
+        XGetErrorText(dpy, ev->error_code, error_buf, 256);
+        cerr << "XError: " << error_buf << " id: " << ev->resourceid << endl;
+#endif // DEBUG
+
+        return 0;
+    }
+}
+
 //! @brief PScreen::Visual constructor.
 //! @param x_visual X Visual to wrap.
 PScreen::PVisual::PVisual(Visual *x_visual) : _x_visual(x_visual),
@@ -84,11 +119,10 @@ PScreen::PVisual::getShiftPrecFromMask(ulong mask, int &shift, int &prec)
 
 //! @brief PScreen constructor
 PScreen::PScreen(Display *dpy, bool honour_randr)
-    : _dpy(dpy), _honour_randr(honour_randr), _fd(-1),
+    : _honour_randr(honour_randr), _fd(-1),
       _screen(-1), _depth(-1),
       _root(None), _visual(0), _colormap(None),
       _modifier_map(0),
-      _num_lock(0), _scroll_lock(0),
       _has_extension_shape(false), _event_shape(-1),
       _has_extension_xinerama(false),
       _has_extension_xrandr(false), _event_xrandr(-1),
@@ -99,6 +133,10 @@ PScreen::PScreen(Display *dpy, bool honour_randr)
     }
     _instance = this;
 
+    XSetErrorHandler(handleXError);
+
+    _dpy = dpy;
+
     XGrabServer(_dpy);
 
     _fd = ConnectionNumber(dpy);
@@ -137,8 +175,7 @@ PScreen::PScreen(Display *dpy, bool honour_randr)
     }
 
     // Figure out what keys the Num and Scroll Locks are
-    _num_lock = getMaskFromKeycode(XKeysymToKeycode(_dpy, XK_Num_Lock));
-    _scroll_lock = getMaskFromKeycode(XKeysymToKeycode(_dpy, XK_Scroll_Lock));
+    setLockKeys();
 
     XSync(_dpy, false);
     XUngrabServer(_dpy);
@@ -155,6 +192,16 @@ PScreen::~PScreen(void) {
     _instance = 0;
 }
 
+/**
+ * Figure out what keys the Num and Scroll Locks are
+ */
+void
+PScreen::setLockKeys(void)
+{
+    _num_lock = getMaskFromKeycode(XKeysymToKeycode(_dpy, XK_Num_Lock));
+    _scroll_lock = getMaskFromKeycode(XKeysymToKeycode(_dpy, XK_Scroll_Lock));
+}
+
 //! @brief Get next event using select to avoid signal blocking
 //! @param ev Event to fill in.
 //! @return true if event was fetched, else false.
@@ -524,6 +571,12 @@ PScreen::updateStrut(void)
           strut->bottom = (*it)->bottom;
         }
     }
+
+    // Update hints on the root window
+    Geometry workarea(_strut.left, _strut.top,
+                      _screen_gm.width - _strut.left - _strut.right, _screen_gm.height - _strut.top - _strut.bottom);
+
+    static_cast<RootWO*>(PWinObj::getRootPWinObj())->setEwmhWorkarea(workarea);
 }
 
 //! @brief Initialize head information
@@ -650,3 +703,7 @@ PScreen::getKeycodeFromMask(uint mask)
 
     return 0;
 }
+
+Display *PScreen::_dpy;
+uint PScreen::_num_lock = 0;
+uint PScreen::_scroll_lock = 0;
