@@ -1,6 +1,6 @@
 //
 // ScreenResources.hh for pekwm
-// Copyright (C) 2004 Claes Nasten <pekdon{@}pekdon{.}net>
+// Copyright (C) 2009 Claes Nasten <pekdon{@}pekdon{.}net>
 //
 // This program is licensed under the GNU GPL.
 // See the LICENSE file for more information.
@@ -36,7 +36,7 @@ ScreenResources::ScreenResources(void)
 #ifdef DEBUG
         cerr << __FILE__ << "@" << __LINE__ << ": "
              << "ScreenResources(" << this << ")::ScreenResources()" << endl
-             << " *** _instance allready set: " << _instance << endl;
+             << " *** _instance already set: " << _instance << endl;
 #endif // DEBUG
     }
     _instance = this;
