@@ -1,6 +1,6 @@
 //
 // StatusWindow.cc for pekwm
-// Copyright (C) 2004-2005 Claes Nasten <pekdon{@}pekdon{.}net>
+// Copyright (C) 2004-2009 Claes Nasten <pekdon{@}pekdon{.}net>
 //
 // This program is licensed under the GNU GPL.
 // See the LICENSE file for more information.
@@ -27,15 +27,15 @@ using std::endl;
 StatusWindow *StatusWindow::_instance = 0;
 
 //! @brief StatusWindow constructor
-StatusWindow::StatusWindow(Display *dpy, Theme *theme)
-    : PDecor(dpy, theme, "STATUSWINDOW"),
+StatusWindow::StatusWindow(Theme *theme)
+    : PDecor(theme, "STATUSWINDOW"),
       _bg(None)
 {
     if (_instance) {
 #ifdef DEBUG
         cerr << __FILE__ << "@" << __LINE__ << ": "
              << "StatusWindow(" << this << ")::StatusWindow()" << endl
-             << " *** _instance allready set: " << _instance << endl;
+             << " *** _instance already set: " << _instance << endl;
 #endif // DEBUG
     }
     _instance = this;
@@ -43,13 +43,13 @@ StatusWindow::StatusWindow(Display *dpy, Theme *theme)
 
     // PWinObj attributes
     _type = PWinObj::WO_STATUS;
-    _layer = LAYER_NONE; // hack, goes over LAYER_MENU
+    setLayer(LAYER_NONE); // hack, goes over LAYER_MENU
     _hidden = true; // don't care about it when changing worskpace etc
 
     XSetWindowAttributes attr;
     attr.event_mask = None;
 
-    _status_wo = new PWinObj(_dpy);
+    _status_wo = new PWinObj;
     _status_wo->setWindow(XCreateWindow(_dpy, _window,
                                         0, 0, 1, 1, 0,
                                         CopyFromParent, CopyFromParent, CopyFromParent,
@@ -86,7 +86,7 @@ StatusWindow::~StatusWindow(void)
 //! @param text Text to draw in StatusWindow
 //! @param do_center Center the StatusWindow on screen. Defaults to false.
 void
-StatusWindow::draw(const std::wstring &text, bool do_center)
+StatusWindow::draw(const std::wstring &text, bool do_center, Geometry *gm)
 {
     uint width, height;
     PFont *font = _theme->getStatusData()->getFont(); // convenience
@@ -103,7 +103,7 @@ StatusWindow::draw(const std::wstring &text, bool do_center)
     }
 
     if (do_center) {
-        center();
+        center(gm);
     }
 
     font->setColor(_theme->getStatusData()->getColor());
@@ -145,11 +145,13 @@ StatusWindow::render(void)
 
 //! @brief Centers the StatusWindow on the current head/screen
 void
-StatusWindow::center(void)
+StatusWindow::center(Geometry *gm)
 {
     Geometry head;
-    PScreen::instance()->getHeadInfo(PScreen::instance()->getCurrHead(), head);
+    if (! gm) {
+        PScreen::instance()->getHeadInfo(PScreen::instance()->getCurrHead(), head);
+        gm = &head;
+    }
 
-    move(head.x + (head.width - _gm.width) / 2,
-         head.y + (head.height - _gm.height) / 2);
+    move(gm->x + (gm->width - _gm.width) / 2, gm->y + (gm->height - _gm.height) / 2);
 }
