@@ -1,6 +1,6 @@
 //
 // Theme.cc for pekwm
-// Copyright © 2003-2008 Claes Nasten <me@pekdon.net>
+// Copyright © 2003-2009 Claes Nasten <me@pekdon.net>
 //
 // This program is licensed under the GNU GPL.
 // See the LICENSE file for more information.
@@ -80,27 +80,31 @@ Theme::PDecorButtonData::load(CfgParser::Entry *section)
         TextureHandler *th = TextureHandler::instance();
         CfgParser::Entry *value;
 
-        value = section->find_entry ("FOCUSED");
+        value = section->find_entry("FOCUSED");
         if (value) {
             _texture[BUTTON_STATE_FOCUSED] = th->getTexture(value->get_value());
         }
         
-        value = section->find_entry ("UNFOCUSED");
+        value = section->find_entry("UNFOCUSED");
         if (value) {
             _texture[BUTTON_STATE_UNFOCUSED] = th->getTexture(value->get_value());
         }
         
-        value = section->find_entry ("PRESSED");
+        value = section->find_entry("PRESSED");
         if (value) {
             _texture[BUTTON_STATE_PRESSED] = th->getTexture(value->get_value());
         }
-        
-        value = section->find_entry ("HOOVER");
+
+        // HOOVER has been kept around due to backwards compatibility.
+        value = section->find_entry("HOVER");
+        if (! value) {
+            value = section->find_entry("HOOVER");
+        }
         if (value) {
-            _texture[BUTTON_STATE_HOOVER] = th->getTexture(value->get_value());
+            _texture[BUTTON_STATE_HOVER] = th->getTexture(value->get_value());
         }
         
-        check ();
+        check();
 
         return true;
     }
@@ -222,9 +226,9 @@ Theme::PDecorData::load(CfgParser::Entry *section)
     list<CfgParserKey*> key_list;
     string value_pad, value_focused, value_unfocused;
 
-    key_list.push_back(new CfgParserKeyInt("HEIGHT", _title_height, 10, 0));
-    key_list.push_back(new CfgParserKeyInt("WIDTHMIN", _title_width_min, 0));
-    key_list.push_back(new CfgParserKeyInt("WIDTHMAX", _title_width_max, 100, 0, 100));
+    key_list.push_back(new CfgParserKeyNumeric<int>("HEIGHT", _title_height, 10, 0));
+    key_list.push_back(new CfgParserKeyNumeric<int>("WIDTHMIN", _title_width_min, 0));
+    key_list.push_back(new CfgParserKeyNumeric<int>("WIDTHMAX", _title_width_max, 100, 0, 100));
     key_list.push_back(new CfgParserKeyBool("WIDTHSYMETRIC", _title_width_symetric));
     key_list.push_back(new CfgParserKeyBool("HEIGHTADAPT", _title_height_adapt));
     key_list.push_back(new CfgParserKeyString("PAD", value_pad, "0 0 0 0", 7));
@@ -356,7 +360,7 @@ Theme::PDecorData::check(void)
 
 //! @brief Loads border data.
 void
-Theme::PDecorData::loadBorder (CfgParser::Entry *section)
+Theme::PDecorData::loadBorder(CfgParser::Entry *section)
 {
     if (! section) {
         return;
@@ -391,7 +395,7 @@ Theme::PDecorData::loadBorder (CfgParser::Entry *section)
 
 //! @brief Loads button data.
 void
-Theme::PDecorData::loadButtons (CfgParser::Entry *section)
+Theme::PDecorData::loadButtons(CfgParser::Entry *section)
 {
     if (! section) {
         return;
@@ -748,8 +752,8 @@ Theme::WorkspaceIndicatorData::load(CfgParser::Entry *section)
     key_list.push_back(new CfgParserKeyString("BACKGROUND", value_tex_bg));
     key_list.push_back(new CfgParserKeyString("WORKSPACE", value_tex_ws));
     key_list.push_back(new CfgParserKeyString("WORKSPACEACTIVE", value_tex_ws_act));
-    key_list.push_back(new CfgParserKeyInt("EDGEPADDING", edge_padding, 5, 0));
-    key_list.push_back(new CfgParserKeyInt("WORKSPACEPADDING", workspace_padding, 2, 0));
+    key_list.push_back(new CfgParserKeyNumeric<int>("EDGEPADDING", edge_padding, 5, 0));
+    key_list.push_back(new CfgParserKeyNumeric<int>("WORKSPACEPADDING", workspace_padding, 2, 0));
 
     section->parse_key_values(key_list.begin(), key_list.end());
     for_each(key_list.begin(), key_list.end(), Util::Free<CfgParserKey*>());  
@@ -813,9 +817,6 @@ Theme::WorkspaceIndicatorData::check(void)
     }
 }
 
-#ifdef HARBOUR
-// HarbourData
-
 /**
  * HarbourData constructor.
  */
@@ -877,13 +878,11 @@ Theme::HarbourData::check(void)
     }
 }
 
-#endif // HARBOUR
-
 // Theme
 
 //! @brief Theme constructor
 Theme::Theme(PScreen *scr)
-    : _scr(scr), _image_handler(0), _theme_mtime(0),
+    : _scr(scr), _image_handler(0),
       _is_loaded(false), _invert_gc(None)
 {
     // image handler
@@ -894,9 +893,7 @@ Theme::Theme(PScreen *scr)
     _section_data_map["STATUS"] = &_status_data;
     _section_data_map["CMDDIALOG"] = &_cmd_d_data;
     _section_data_map["WORKSPACEINDICATOR"] = &_workspace_indicator_data;
-#ifdef HARBOUR
     _section_data_map["HARBOUR"] = &_harbour_data;
-#endif // HARBOUR
 
     // window gc's
     XGCValues gv;
@@ -906,7 +903,6 @@ Theme::Theme(PScreen *scr)
     _invert_gc = XCreateGC(_scr->getDpy(), _scr->getRoot(),
                            GCFunction|GCSubwindowMode|GCLineWidth, &gv);
 
-
     _scr->grabServer();
 
     load(Config::instance()->getThemeFile());
@@ -936,7 +932,7 @@ Theme::load(const std::string &dir)
     }
     string theme_file(norm_dir + string("theme"));
 
-    if (! Util::requireReload(_theme_path, theme_file, _theme_mtime)) {
+    if (! Util::requireReload(_cfg_state, theme_file)) {
         return false;
     }
 
@@ -955,8 +951,8 @@ Theme::load(const std::string &dir)
 
     if (! theme.parse(theme_file)) {
         _theme_dir = DATADIR "/pekwm/themes/default/";
-        _theme_path = _theme_dir + string("theme");
-        if (! theme.parse(_theme_path)) {
+        theme_file = _theme_dir + string("theme");
+        if (! theme.parse(theme_file)) {
             cerr << " *** WARNING: couldn't load " << _theme_dir << " or default theme." << endl;
             theme_ok = false;
         }
@@ -964,11 +960,17 @@ Theme::load(const std::string &dir)
 
     // Setup quirks and requirements before parsing.
     if (theme_ok) {
-        loadThemeRequire(theme);
+        if (theme.is_dynamic_content()) {
+            _cfg_state.clear();
+        } else {
+            _cfg_state = theme.get_file_list();
+        }
+        loadThemeRequire(theme, theme_file);
     }
 
     // Set image basedir.
-    _image_handler->setDir(_theme_dir);
+    _image_handler->path_clear();
+    _image_handler->path_push_back(_theme_dir);
 
     // Load decor data.
     CfgParser::Entry *section = theme.get_entry_root()->find_section("PDECOR");
@@ -1012,7 +1014,7 @@ Theme::load(const std::string &dir)
  * Load template quirks.
  */
 void
-Theme::loadThemeRequire(CfgParser &theme_cfg)
+Theme::loadThemeRequire(CfgParser &theme_cfg, std::string &file)
 {
     CfgParser::Entry *section;
 
@@ -1028,9 +1030,8 @@ Theme::loadThemeRequire(CfgParser &theme_cfg)
 
         // Re-load configuration with templates enabled.
         if (value_templates) {
-            string theme_path(_theme_path);
             theme_cfg.clear(true);
-            theme_cfg.parse(theme_path, CfgParserSource::SOURCE_FILE, true);
+            theme_cfg.parse(file, CfgParserSource::SOURCE_FILE, true);
         }
     }
 }
