@@ -1,6 +1,6 @@
 //
 // Workspaces.cc for pekwm
-// Copyright © 2002-2008 Claes Nasten <me@pekdon.net>
+// Copyright © 2002-2009 Claes Nasten <me@pekdon.net>
 //
 // This program is licensed under the GNU GPL.
 // See the LICENSE file for more information.
@@ -19,6 +19,8 @@
 #include "PDecor.hh"
 #include "Frame.hh"
 #include "Client.hh" // For isSkip()
+#include "WindowManager.hh"
+#include "WorkspaceIndicator.hh"
 
 #include <iostream>
 #include <sstream>
@@ -42,8 +44,8 @@ using std::endl;
 
 // Workspaces::Workspace
 
-Workspaces::Workspace::Workspace(const std::wstring &name, uint number, const std::list<PWinObj*> &wo_list)
-  : _name(name), _number(number), _wo_list(wo_list), _last_focused(0)
+Workspaces::Workspace::Workspace(const std::wstring &name, uint number)
+  : _name(name), _number(number), _last_focused(0)
 {
 }
 
@@ -56,7 +58,7 @@ Workspaces::Workspace::~Workspace(void)
 Workspaces *Workspaces::_instance = 0;
 
 
-//! @brief Workpsaces constructor
+//! @brief Workspaces constructor
 Workspaces::Workspaces(uint number, uint per_row)
   : _active(0), _previous(0), _per_row(per_row)
 {
@@ -64,7 +66,7 @@ Workspaces::Workspaces(uint number, uint per_row)
     if (_instance) {
         cerr << __FILE__ << "@" << __LINE__ << ": "
              << "Workspaces(" << this << ")::Workspaces(" << number << ")"
-             << endl << " *** _instance allready set: " << _instance << endl;
+             << endl << " *** _instance already set: " << _instance << endl;
     }
 #endif // DEBUG
     _instance = this;
@@ -80,7 +82,7 @@ Workspaces::Workspaces(uint number, uint per_row)
 
     // create new workspaces
     for (uint i = 0; i < number; ++i) {
-      _workspace_list.push_back(new Workspace(getWorkspaceName(i), i, _wo_list));
+        _workspace_list.push_back(new Workspace(getWorkspaceName(i), i));
     }
 }
 
@@ -131,14 +133,14 @@ Workspaces::setSize(uint number)
 
     } else { // We need more workspaces, lets create some
         for (uint i = before; i < number; ++i) {
-          _workspace_list.push_back(new Workspace(getWorkspaceName(i), i, _wo_list));
+            _workspace_list.push_back(new Workspace(getWorkspaceName(i), i));
         }
     }
 
     // Tell the rest of the world how many workspaces we have.
     XChangeProperty(PScreen::instance()->getDpy(),
                     PScreen::instance()->getRoot(),
-                    EwmhAtoms::instance()->getAtom(NET_NUMBER_OF_DESKTOPS),
+                    Atoms::getAtom(NET_NUMBER_OF_DESKTOPS),
                     XA_CARDINAL, 32, PropModeReplace,
                     (uchar *) &number, 1);
 
@@ -162,7 +164,7 @@ Workspaces::setNames(void)
 
 //! @brief Activates Workspace workspace and sets the right hints
 //! @param num Workspace to activate
-//! @param focus wheter or not to focus a window after switch
+//! @param focus whether or not to focus a window after switch
 void
 Workspaces::setWorkspace(uint num, bool focus)
 {
@@ -190,11 +192,18 @@ Workspaces::setWorkspace(uint num, bool focus)
     // switch workspace
     hideAll(_active);
     AtomUtil::setLong(PScreen::instance()->getRoot(),
-                      EwmhAtoms::instance()->getAtom(NET_CURRENT_DESKTOP),
+                      Atoms::getAtom(NET_CURRENT_DESKTOP),
                       num);
     unhideAll(num, focus);
 
     PScreen::instance()->ungrabServer(true);
+
+    // Show workspace indicator if requested
+    if (Config::instance()->getShowWorkspaceIndicator() > 0) {
+        WindowManager::instance()->getWorkspaceIndicator()->render();
+        WindowManager::instance()->getWorkspaceIndicator()->mapWindowRaised();
+        WindowManager::instance()->getWorkspaceIndicator()->updateHideTimer(Config::instance()->getShowWorkspaceIndicator());
+    }
 }
 
 //! @brief
@@ -300,22 +309,22 @@ Workspaces::warpToWorkspace(uint num, int dir)
     if (dir != 0) {
       switch(dir) {
       case 1:
-        x = PScreen::instance()->getWidth() - std::max(Config::instance()->getScreenEdgeSize(SCREEN_EDGE_RIGHT) * 2, 2);
+        x = PScreen::instance()->getWidth() - std::max(Config::instance()->getScreenEdgeSize(SCREEN_EDGE_LEFT) + 2, 2);
         break;
       case 2:
         x = std::max(Config::instance()->getScreenEdgeSize(SCREEN_EDGE_RIGHT) * 2, 2);
         break;
       case -1:
-        y = PScreen::instance()->getHeight() - std::max(Config::instance()->getScreenEdgeSize(SCREEN_EDGE_BOTTOM) * 2, 2);;
+        y = PScreen::instance()->getHeight() - std::max(Config::instance()->getScreenEdgeSize(SCREEN_EDGE_BOTTOM) + 2, 2);
         break;
       case -2:
-        y = std::max(Config::instance()->getScreenEdgeSize(SCREEN_EDGE_TOP) * 2, 2);
+        y = std::max(Config::instance()->getScreenEdgeSize(SCREEN_EDGE_TOP) + 2, 2);
         break;
       }
 
         // warp pointer
-        XWarpPointer(PScreen::instance()->getDpy(), None, PScreen::instance()->getRoot(),
-                     0, 0, 0, 0, x, y);
+        XWarpPointer(PScreen::instance()->getDpy(), None,
+                     PScreen::instance()->getRoot(), 0, 0, 0, 0, x, y);
     }
 
     // set workpsace
@@ -324,32 +333,36 @@ Workspaces::warpToWorkspace(uint num, int dir)
     return true;
 }
 
-//! @brief Adds a PWinObj to the stacking list.
-//! @param wo PWinObj to insert
-//! @param raise Defaults to true, wheter to check for bottom or top.
+/**
+ * Adds a PWinObj to the stacking list.
+ * @param wo PWinObj to insert
+ * @param raise Whether to insert at the bottom or top of the layer (defaults to true).
+ */
 void
-Workspaces::insert(PWinObj* wo, bool raise)
+Workspaces::insert(PWinObj *wo, bool raise)
 {
-    bool inserted = false;
-
-    list<PWinObj*>::iterator it(_wo_list.begin());
-    for (; it != _wo_list.end() && ! inserted; ++it) {
-        if (raise
-                ? (wo->getLayer() < (*it)->getLayer())
-                : (wo->getLayer() <= (*it)->getLayer())) {
-            _wo_list.insert(it, wo);
-
-            stackWinUnderWin((*it)->getWindow(), wo->getWindow());
-
-            inserted = true;
+    list<PWinObj*>::iterator it(_wo_list.begin()), position(_wo_list.end());
+    for (; it != _wo_list.end() && position == _wo_list.end(); ++it) {
+        if (raise) {
+            // If raising, make sure the inserted wo gets below the first
+            // window in the next layer.
+            if ((*it)->getLayer() > wo->getLayer()) {
+                position = it;
+            }
+        } else {
+            // If lowering, put the window below the first window with the same level.
+            if (wo->getLayer() <= (*it)->getLayer()) {
+                position = it;
+            }
         }
     }
 
-    if (! inserted) {
-        _wo_list.push_back(wo);
-        // we can't do raise here as it'll end up in an recursive loop
-        // if we use Workspaces::raise to restack the PWinObj.
-        XRaiseWindow(PScreen::instance()->getDpy(), wo->getWindow());
+    _wo_list.insert(position, wo);
+
+    if (position == _wo_list.end()) {
+         XRaiseWindow(PScreen::instance()->getDpy(), wo->getWindow());
+    } else {
+         stackWinUnderWin((*position)->getWindow(), wo->getWindow());
     }
 }
 
@@ -439,8 +452,9 @@ Workspaces::raise(PWinObj* wo)
 {
     list<PWinObj*>::iterator it(find(_wo_list.begin(), _wo_list.end(), wo));
 
-    if (it == _wo_list.end()) // no Frame to raise.
+    if (it == _wo_list.end()) { // no Frame to raise.
         return;
+    }
     _wo_list.erase(it);
 
     insert(wo, true); // reposition and restack
@@ -477,8 +491,9 @@ Workspaces::stackAbove(PWinObj *wo, Window win, bool restack)
 
                 // Before restacking make sure we are the active frame
                 // also that there are two different frames
-                if (restack)
-                    stackWinUnderWin((*--it)->getWindow(), wo->getWindow());
+                if (restack) {
+                    stackWinUnderWin(win, wo->getWindow());
+                }
                 break;
             }
         }
@@ -501,8 +516,9 @@ Workspaces::stackBelow(PWinObj* wo, Window win, bool restack)
                 _wo_list.erase(old_pos);
                 _wo_list.insert(it, wo);
 
-                if (restack)
-                    stackWinUnderWin((*it)->getWindow(), wo->getWindow());
+                if (restack) {
+                    stackWinUnderWin(wo->getWindow(), win);
+                }
                 break;
             }
         }
@@ -540,7 +556,7 @@ Workspaces::stackWinUnderWin(Window over, Window under)
     if (over == under) {
         return;
     }
-    
+
     Window windows[2] = { over, under };
     XRestackWindows(PScreen::instance()->getDpy(), windows, 2);
 }
@@ -574,41 +590,79 @@ Workspaces::getTopWO(uint type_mask)
     return 0;
 }
 
-//! @brief Updates the Ewmh Client list and Stacking list hint.
-void
-Workspaces::updateClientStackingList(bool client, bool stacking)
+/**
+ * Builds a list of all clients in stacking order, clients in the same
+ * frame come after each other.
+ */
+Window*
+Workspaces::buildClientList(unsigned int &num_windows)
 {
-    if (! Frame::frame_size()) {
-        return;
+    Frame *frame;
+    Client *client, *client_active;
+
+    list<Window> windows_list;
+    list<PWinObj*>::iterator it_f, it_c;
+    for (it_f = _wo_list.begin(); it_f != _wo_list.end(); ++it_f) {
+        if ((*it_f)->getType() != PWinObj::WO_FRAME) {
+	  continue;
+	}
+
+	frame = static_cast<Frame*>(*it_f);
+	client_active = frame->getActiveClient();
+
+	if (Config::instance()->isReportAllClients()) {
+            for (it_c = frame->begin(); it_c != frame->end(); ++it_c) {
+                client = dynamic_cast<Client*>(*it_c);
+                if (client
+		    && ! client->isSkip(SKIP_TASKBAR)
+		    && client != client_active) {
+                    windows_list.push_back(client->getWindow());
+	        }
+	    }
+	}
+
+	if (client_active && ! client_active->isSkip(SKIP_TASKBAR)) {
+	    windows_list.push_back(client_active->getWindow());
+	}
     }
 
-    Client *cl;
-
-    // Create array for holding windows, can not be more than this.
-    Window *windows = new Window[Frame::frame_size()];
-
-    // Copy Frame windows in order
-    uint pos = 0;
-    list<PWinObj*>::iterator it(_wo_list.begin());
-    for (; it != _wo_list.end(); ++it) {
-        if ((*it)->getType() == PWinObj::WO_FRAME) {
-            cl = static_cast<Client*>(static_cast<Frame*>((*it))->getActiveChild());
-            if (cl && ! cl->isSkip(SKIP_TASKBAR)) {
-                windows[pos++] = cl->getWindow();
-            }
-        }
+    num_windows = windows_list.size();
+    Window *windows = new Window[num_windows ? num_windows : 1];
+    if (num_windows > 0) {
+      copy(windows_list.begin(), windows_list.end(), windows);
     }
 
-    if (client) {
-        AtomUtil::setWindows(PScreen::instance()->getRoot(),
-                             EwmhAtoms::instance()->getAtom(NET_CLIENT_LIST),
-                             windows, pos);
-    }
-    if (stacking) {
-        AtomUtil::setWindows(PScreen::instance()->getRoot(),
-                             EwmhAtoms::instance()->getAtom(NET_CLIENT_LIST_STACKING),
-                             windows, pos);
-    }
+    return windows;
+}
+
+/**
+ * Updates the Ewmh Client list hint.
+ */
+void
+Workspaces::updateClientList(void)
+{
+    unsigned int num_windows;
+    Window *windows = buildClientList(num_windows);
+
+    AtomUtil::setWindows(PScreen::instance()->getRoot(),
+			 Atoms::getAtom(NET_CLIENT_LIST),
+			 windows, num_windows);
+    
+    delete [] windows;
+}
+
+/**
+ * Updates the Ewmh Stacking list hint.
+ */
+void
+Workspaces::updateClientStackingList(void)
+{
+    unsigned int num_windows;
+    Window *windows = buildClientList(num_windows);
+
+    AtomUtil::setWindows(PScreen::instance()->getRoot(),
+			 Atoms::getAtom(NET_CLIENT_LIST_STACKING),
+			 windows, num_windows);
 
     delete [] windows;
 }
@@ -647,6 +701,29 @@ Workspaces::placeWo(PWinObj *wo, Window parent)
     }
 }
 
+/**
+ * Make sure window is inside screen boundaries.
+ */
+void
+Workspaces::placeWoInsideScreen(PWinObj *wo)
+{
+    Geometry gm_before(wo->getX(), wo->getY(), wo->getWidth(), wo->getHeight());
+    Geometry gm_after(gm_before);
+
+    Strut *strut = 0;
+    if (wo->getType() == PWinObj::WO_FRAME) {
+        Client *client = static_cast<Frame*>(wo)->getActiveClient();
+        if (client) {
+            strut = client->getStrut();
+        }
+    }
+
+    placeInsideScreen(gm_after, strut);
+    if (gm_before != gm_after) {
+        wo->move(gm_after.x, gm_after.y);
+    }
+}
+
 //! @brief Tries to find empty space to place the client in
 //! @return true if client got placed, else false
 //! @todo What should we do about Xinerama as when we don't have it enabled we care about the struts.
@@ -821,11 +898,17 @@ Workspaces::placeCenteredOnParent(PWinObj *wo, Window parent)
 
 //! @brief Makes sure the window is inside the screen.
 void
-Workspaces::placeInsideScreen(Geometry &gm)
+Workspaces::placeInsideScreen(Geometry &gm, Strut *strut)
 {
+    // Do not include screen edges when calculating the position if the window
+    // has a strut as it then is likely to be a panel or the like placed
+    // along the edge of the screen.
     Geometry head;
-    PScreen::instance()->getHeadInfoWithEdge(PScreen::instance()->getCurrHead(),
-            head);
+    if (strut) {
+        PScreen::instance()->getHeadInfo(PScreen::instance()->getCurrHead(), head);
+    } else {
+        PScreen::instance()->getHeadInfoWithEdge(PScreen::instance()->getCurrHead(), head);
+    }
 
     if (gm.x < head.x) {
         gm.x = head.x;
@@ -858,6 +941,17 @@ Workspaces::isEmptySpace(int x, int y, const PWinObj* wo)
             continue;
         }
 
+        // Also skip windows tagged as Maximized as they cause us to
+        // automatically fail.
+        if ((*it)->getType() == PWinObj::WO_FRAME) {
+            Client *client = static_cast<Frame*>((*it))->getActiveClient();
+            if (client &&
+                (client->isFullscreen()
+                 || (client->isMaximizedVert() && client->isMaximizedHorz()))) {
+                continue;
+            }
+        }
+
         // Check if we are "intruding" on some other window's place
         if (((*it)->getX() < signed(x + wo->getWidth())) &&
             (signed((*it)->getX() + (*it)->getWidth()) > x) &&
